"""Timeseries data resources"""
import flask
from flask_smorest import abort

from bemserver_core.model import Timeseries, TimeseriesDataState, Campaign
from bemserver_core.input_output import tsdcsvio, tsdjsonio
from bemserver_core.exceptions import (
    TimeseriesNotFoundError,
    TimeseriesDataIOError,
)

from bemserver_api import Blueprint

from .schemas import (
    TimeseriesDataGetByIDQueryArgsSchema,
    TimeseriesDataDeleteByIDQueryArgsSchema,
    TimeseriesDataGetByNameQueryArgsSchema,
    TimeseriesDataDeleteByNameQueryArgsSchema,
    TimeseriesDataGetByIDAggregateQueryArgsSchema,
    TimeseriesDataGetByNameAggregateQueryArgsSchema,
    TimeseriesDataPostQueryArgsSchema,
    TimeseriesDataGetAcceptHeader,
)


def _get_data_state(data_state_id):
    return TimeseriesDataState.get_by_id(data_state_id) or abort(
        422, errors={"query": {"data_state": "Unknown data state ID"}}
    )


def _get_many_timeseries_by_id(timeseries_ids):
    try:
        return Timeseries.get_many_by_id(timeseries_ids)
    except TimeseriesNotFoundError as exc:
        abort(422, message=str(exc))


def _get_many_timeseries_by_name(campaign, timeseries_ids):
    try:
        return Timeseries.get_many_by_name(campaign, timeseries_ids)
    except TimeseriesNotFoundError as exc:
        abort(422, message=str(exc))


blp = Blueprint(
    "TimeseriesData",
    __name__,
    url_prefix="/timeseries_data",
    description="Operations on timeseries data",
)


blp4c = Blueprint(
    "TimeseriesDataForCampaign",
    __name__,
    url_prefix="/timeseries_data/campaign/<int:campaign_id>",
    description="Operations on timeseries data for a given campaign",
)


@blp.route("/", methods=("GET",))
@blp.login_required
@blp.arguments(TimeseriesDataGetByIDQueryArgsSchema, location="query")
@blp.arguments(TimeseriesDataGetAcceptHeader, location="headers", error_status_code=406)
@blp.response(200, content_type="application/json")
@blp.alt_response(200, content_type="application/csv")
def get(args, headers):
    """Get timeseries data

    Returns data in either JSON or CSV format.

    JSON: each key is a timestamp ID as string. For each timeseries, values are
    passed a {timestamp: value} mappings.

    ```
    {
        "1": {
            "2020-01-01T00:00:00+00:00": 0.1,
            "2020-01-01T10:00:00+00:00": 0.2,
            "2020-01-01T20:00:00+00:00": 0.3,
        },
        "2": {
            "2020-01-01T00:00:00+00:00": 1.1,
            "2020-01-01T10:00:00+00:00": 1.2,
            "2020-01-01T20:00:00+00:00": 1.3,
        },
        "3": {
            "2020-01-01T00:00:00+00:00": 2.1,
            "2020-01-01T10:00:00+00:00": 2.2,
            "2020-01-01T20:00:00+00:00": 2.3,
        },
    }
    ```
    CSV: the first column is the timestamp as timezone aware datetime and each
    other column is a timeseries data. Column headers are timeseries IDs.

    ```
    Datetime,1,2,3
    2020-01-01T00:00:00+00:00,0.1,1.1,2.1
    2020-01-01T00:10:00+00:00,0.2,1.2,2.2
    2020-01-01T00:20:00+00:00,0.3,1.3,2.3
    ```
    """
    mime_type = headers["accept"]

    timeseries = _get_many_timeseries_by_id(args["timeseries"])
    data_state = _get_data_state(args["data_state"])

    if mime_type == "application/csv":
        resp = tsdcsvio.export_csv(
            args["start_time"],
            args["end_time"],
            timeseries,
            data_state,
            timezone=args["timezone"],
            col_label="id",
        )
    else:
        resp = tsdjsonio.export_json(
            args["start_time"],
            args["end_time"],
            timeseries,
            data_state,
            timezone=args["timezone"],
            col_label="id",
        )

    return flask.Response(resp, mimetype=mime_type)


@blp.route("/aggregate", methods=("GET",))
@blp.login_required
@blp.arguments(TimeseriesDataGetByIDAggregateQueryArgsSchema, location="query")
@blp.arguments(TimeseriesDataGetAcceptHeader, location="headers", error_status_code=406)
@blp.response(200, content_type="application/csv")
def get_aggregate(args, headers):
    """Get aggregated timeseries data as CSV file

    Returns data in either JSON or CSV format.

    JSON: each key is a timestamp ID as string. For each timeseries, values are
    passed a {timestamp: value} mappings.

    ```
    {
        "1": {
            "2020-01-01T00:00:00+00:00": 0.1,
            "2020-01-01T10:00:00+00:00": 0.2,
            "2020-01-01T20:00:00+00:00": 0.3,
        },
        "2": {
            "2020-01-01T00:00:00+00:00": 1.1,
            "2020-01-01T10:00:00+00:00": 1.2,
            "2020-01-01T20:00:00+00:00": 1.3,
        },
        "3": {
            "2020-01-01T00:00:00+00:00": 2.1,
            "2020-01-01T10:00:00+00:00": 2.2,
            "2020-01-01T20:00:00+00:00": 2.3,
        },
    }
    ```

    CSV: the first column is the timestamp as timezone aware datetime and each
    other column is a timeseries data. Column headers are timeseries IDs.

    ```
    Datetime,1,2,3
    2020-01-01T00:00:00+00:00,0.1,1.1,2.1
    2020-01-01T00:10:00+00:00,0.2,1.2,2.2
    2020-01-01T00:20:00+00:00,0.3,1.3,2.3
    ```
    """
    mime_type = headers["accept"]

    timeseries = _get_many_timeseries_by_id(args["timeseries"])
    data_state = _get_data_state(args["data_state"])

    if mime_type == "application/csv":
        resp = tsdcsvio.export_csv_bucket(
            args["start_time"],
            args["end_time"],
            timeseries,
            data_state,
            args["bucket_width_value"],
            args["bucket_width_unit"],
            args["aggregation"],
            timezone=args["timezone"],
            col_label="id",
        )
    else:
        resp = tsdjsonio.export_json_bucket(
            args["start_time"],
            args["end_time"],
            timeseries,
            data_state,
            args["bucket_width_value"],
            args["bucket_width_unit"],
            args["aggregation"],
            timezone=args["timezone"],
            col_label="id",
        )

    return flask.Response(resp, mimetype=mime_type)


@blp.route("/", methods=("POST",))
@blp.login_required
@blp.arguments(TimeseriesDataPostQueryArgsSchema, location="query")
@blp.arguments(TimeseriesDataGetAcceptHeader, location="headers", error_status_code=406)
@blp.response(201)
def post(args, headers):
    """Post timeseries data

    Loads data in either JSON or CSV format.

    JSON: each key is a timestamp ID as string. For each timeseries, values are
    passed a {timestamp: value} mappings.

    ```
    {
        "1": {
            "2020-01-01T00:00:00+00:00": 0.1,
            "2020-01-01T10:00:00+00:00": 0.2,
            "2020-01-01T20:00:00+00:00": 0.3,
        },
        "2": {
            "2020-01-01T00:00:00+00:00": 1.1,
            "2020-01-01T10:00:00+00:00": 1.2,
            "2020-01-01T20:00:00+00:00": 1.3,
        },
        "3": {
            "2020-01-01T00:00:00+00:00": 2.1,
            "2020-01-01T10:00:00+00:00": 2.2,
            "2020-01-01T20:00:00+00:00": 2.3,
        },
    }
    ```

    CSV: the first column is the timestamp as timezone aware datetime and each
    other column is a timeseries data. Column headers are timeseries IDs.

    ```
    Datetime,1,2,3
    2020-01-01T00:00:00+00:00,0.1,1.1,2.1
    2020-01-01T00:10:00+00:00,0.2,1.2,2.2
    2020-01-01T00:20:00+00:00,0.3,1.3,2.3
    ```
    """
    mime_type = headers["accept"]

    data_state = _get_data_state(args["data_state"])

    try:
        data = flask.request.get_data(cache=True).decode("utf-8")
    except UnicodeDecodeError as exc:
        abort(422, message=str(exc))

    try:
        if mime_type == "application/csv":
            tsdcsvio.import_csv(data, data_state)
        else:
            tsdjsonio.import_json(data, data_state)
    except (TimeseriesNotFoundError, TimeseriesDataIOError) as exc:
        abort(422, message=str(exc))


@blp.route("/", methods=("DELETE",))
@blp.login_required
@blp.arguments(TimeseriesDataDeleteByIDQueryArgsSchema, location="query")
@blp.response(204)
def delete(args):
    """Delete timeseries data"""
    timeseries = _get_many_timeseries_by_id(args["timeseries"])
    data_state = _get_data_state(args["data_state"])

    tsdcsvio.delete(
        args["start_time"],
        args["end_time"],
        timeseries,
        data_state,
    )


@blp4c.route("/", methods=("GET",))
@blp4c.login_required
@blp4c.arguments(TimeseriesDataGetByNameQueryArgsSchema, location="query")
@blp4c.arguments(
    TimeseriesDataGetAcceptHeader, location="headers", error_status_code=406
)
@blp4c.response(200, content_type="application/csv")
def get_for_campaign(args, headers, campaign_id):
    """Get timeseries data for a given campaign

    Returns data in either JSON or CSV format.

    JSON: each key is a timestamp name as string. For each timeseries, values
    are passed a {timestamp: value} mappings.

    ```
    {
        "Timeseries 1": {
            "2020-01-01T00:00:00+00:00": 0.1,
            "2020-01-01T10:00:00+00:00": 0.2,
            "2020-01-01T20:00:00+00:00": 0.3,
        },
        "Timeseries 2": {
            "2020-01-01T00:00:00+00:00": 1.1,
            "2020-01-01T10:00:00+00:00": 1.2,
            "2020-01-01T20:00:00+00:00": 1.3,
        },
        "Timeseries 3": {
            "2020-01-01T00:00:00+00:00": 2.1,
            "2020-01-01T10:00:00+00:00": 2.2,
            "2020-01-01T20:00:00+00:00": 2.3,
        },
    }
    ```

    CSV: the first column is the timestamp as timezone aware datetime and each
    other column is a timeseries data. Column headers are timeseries names.

    ```
    Datetime,Timeseries 1,Timeseries 2,Timeseries 3
    2020-01-01T00:00:00+00:00,0.1,1.1,2.1
    2020-01-01T00:10:00+00:00,0.2,1.2,2.2
    2020-01-01T00:20:00+00:00,0.3,1.3,2.3
    ```
    """
    mime_type = headers["accept"]

    campaign = Campaign.get_by_id(campaign_id) or abort(404)
    timeseries = _get_many_timeseries_by_name(campaign, args["timeseries"])
    data_state = _get_data_state(args["data_state"])

    if mime_type == "application/csv":
        resp = tsdcsvio.export_csv(
            args["start_time"],
            args["end_time"],
            timeseries,
            data_state,
            timezone=args["timezone"],
            col_label="name",
        )
    else:
        resp = tsdjsonio.export_json(
            args["start_time"],
            args["end_time"],
            timeseries,
            data_state,
            timezone=args["timezone"],
            col_label="name",
        )

    return flask.Response(resp, mimetype=mime_type)


@blp4c.route("/aggregate", methods=("GET",))
@blp4c.login_required
@blp4c.arguments(TimeseriesDataGetByNameAggregateQueryArgsSchema, location="query")
@blp4c.arguments(
    TimeseriesDataGetAcceptHeader, location="headers", error_status_code=406
)
@blp4c.response(200, content_type="application/csv")
def get_aggregate_for_campaign(args, headers, campaign_id):
    """Get aggregated timeseries data for a given campaign

    Returns data in either JSON or CSV format.

    JSON: each key is a timestamp name as string. For each timeseries, values
    are passed a {timestamp: value} mappings.

    ```
    {
        "Timeseries 1": {
            "2020-01-01T00:00:00+00:00": 0.1,
            "2020-01-01T10:00:00+00:00": 0.2,
            "2020-01-01T20:00:00+00:00": 0.3,
        },
        "Timeseries 2": {
            "2020-01-01T00:00:00+00:00": 1.1,
            "2020-01-01T10:00:00+00:00": 1.2,
            "2020-01-01T20:00:00+00:00": 1.3,
        },
        "Timeseries 3": {
            "2020-01-01T00:00:00+00:00": 2.1,
            "2020-01-01T10:00:00+00:00": 2.2,
            "2020-01-01T20:00:00+00:00": 2.3,
        },
    }
    ```

    CSV: the first column is the timestamp as timezone aware datetime and each
    other column is a timeseries data. Column headers are timeseries names.

    ```
    Datetime,Timeseries 1,Timeseries 2,Timeseries 3
    2020-01-01T00:00:00+00:00,0.1,1.1,2.1
    2020-01-01T00:10:00+00:00,0.2,1.2,2.2
    2020-01-01T00:20:00+00:00,0.3,1.3,2.3
    ```
    """
    mime_type = headers["accept"]

    campaign = Campaign.get_by_id(campaign_id) or abort(404)
    timeseries = _get_many_timeseries_by_name(campaign, args["timeseries"])
    data_state = _get_data_state(args["data_state"])

    if mime_type == "application/csv":
        resp = tsdcsvio.export_csv_bucket(
            args["start_time"],
            args["end_time"],
            timeseries,
            data_state,
            args["bucket_width_value"],
            args["bucket_width_unit"],
            args["aggregation"],
            timezone=args["timezone"],
            col_label="name",
        )
    else:
        resp = tsdjsonio.export_json_bucket(
            args["start_time"],
            args["end_time"],
            timeseries,
            data_state,
            args["bucket_width_value"],
            args["bucket_width_unit"],
            args["aggregation"],
            timezone=args["timezone"],
            col_label="name",
        )

    return flask.Response(resp, mimetype=mime_type)


@blp4c.route("/", methods=("POST",))
@blp4c.login_required
@blp4c.arguments(TimeseriesDataPostQueryArgsSchema, location="query")
@blp4c.arguments(
    TimeseriesDataGetAcceptHeader, location="headers", error_status_code=406
)
@blp4c.response(201)
def post_for_campaign(args, headers, campaign_id):
    """Post timeseries data for a given campaign

    Loads data in either JSON or CSV format.

    JSON: each key is a timestamp ID as string. For each timeseries, values are
    passed a {timestamp: value} mappings.

    ```
    {
        "1": {
            "2020-01-01T00:00:00+00:00": 0.1,
            "2020-01-01T10:00:00+00:00": 0.2,
            "2020-01-01T20:00:00+00:00": 0.3,
        },
        "2": {
            "2020-01-01T00:00:00+00:00": 1.1,
            "2020-01-01T10:00:00+00:00": 1.2,
            "2020-01-01T20:00:00+00:00": 1.3,
        },
        "3": {
            "2020-01-01T00:00:00+00:00": 2.1,
            "2020-01-01T10:00:00+00:00": 2.2,
            "2020-01-01T20:00:00+00:00": 2.3,
        },
    }
    ```

    CSV: the first column is the timestamp as timezone aware datetime and each
    other column is a timeseries data. Column headers are timeseries IDs.

    ```
    Datetime,1,2,3
    2020-01-01T00:00:00+00:00,0.1,1.1,2.1
    2020-01-01T00:10:00+00:00,0.2,1.2,2.2
    2020-01-01T00:20:00+00:00,0.3,1.3,2.3
    ```
    """
    mime_type = headers["accept"]

    campaign = Campaign.get_by_id(campaign_id) or abort(404)
    data_state = _get_data_state(args["data_state"])

    try:
        data = flask.request.get_data(cache=True).decode("utf-8")
    except UnicodeDecodeError as exc:
        abort(422, message=str(exc))

    try:
        if mime_type == "application/csv":
            tsdcsvio.import_csv(data, data_state, campaign=campaign)
        else:
            tsdjsonio.import_json(data, data_state, campaign=campaign)
    except (TimeseriesNotFoundError, TimeseriesDataIOError) as exc:
        abort(422, message=str(exc))


@blp4c.route("/", methods=("DELETE",))
@blp4c.login_required
@blp4c.arguments(TimeseriesDataDeleteByNameQueryArgsSchema, location="query")
@blp4c.response(204)
def delete_for_campaign(args, campaign_id):
    """Delete timeseries data for a given campaign"""
    campaign = Campaign.get_by_id(campaign_id) or abort(404)
    timeseries = _get_many_timeseries_by_name(campaign, args["timeseries"])
    data_state = _get_data_state(args["data_state"])

    tsdcsvio.delete(
        args["start_time"],
        args["end_time"],
        timeseries,
        data_state,
    )



"""Storey resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import Storey

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import StoreySchema, StoreyPutSchema, StoreyQueryArgsSchema


blp = Blueprint(
    "Storey", __name__, url_prefix="/storeys", description="Operations on storeys"
)


@blp.route("/")
class StoreyViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(StoreyQueryArgsSchema, location="query")
    @blp.response(200, StoreySchema(many=True))
    def get(self, args):
        """List storeys"""
        return Storey.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(StoreySchema)
    @blp.response(201, StoreySchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new storey"""
        item = Storey.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class StoreyByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, StoreySchema)
    def get(self, item_id):
        """Get storey by ID"""
        item = Storey.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(StoreyPutSchema)
    @blp.response(200, StoreySchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing storey"""
        item = Storey.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, StoreySchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a storey"""
        item = Storey.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, StoreySchema)
        item.delete()
        db.session.commit()



"""Structural element properties resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import StructuralElementProperty

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import (
    StructuralElementPropertySchema,
    StructuralElementPropertyPutSchema,
    StructuralElementPropertyQueryArgsSchema,
)


blp = Blueprint(
    "StructuralElementProperty",
    __name__,
    url_prefix="/structural_element_properties",
    description="Operations on structural element properties",
)


@blp.route("/")
class StructuralElementPropertiesViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(StructuralElementPropertyQueryArgsSchema, location="query")
    @blp.response(200, StructuralElementPropertySchema(many=True))
    def get(self, args):
        """List structural element properties"""
        return StructuralElementProperty.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(StructuralElementPropertySchema)
    @blp.response(201, StructuralElementPropertySchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new structural element property"""
        item = StructuralElementProperty.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class StructuralElementPropertyByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, StructuralElementPropertySchema)
    def get(self, item_id):
        """Get structural element property by ID"""
        item = StructuralElementProperty.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(StructuralElementPropertyPutSchema)
    @blp.response(200, StructuralElementPropertySchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing structural element property"""
        item = StructuralElementProperty.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, StructuralElementPropertySchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a structural element property"""
        item = StructuralElementProperty.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, StructuralElementPropertySchema)
        item.delete()
        db.session.commit()



"""Completeness resources"""
from flask_smorest import abort

from bemserver_core.process.completeness import compute_completeness
from bemserver_core.model import Timeseries, TimeseriesDataState
from bemserver_core.exceptions import TimeseriesNotFoundError


from bemserver_api import Blueprint

from .schemas import CompletenessSchema, CompletenessQueryArgsSchema


blp = Blueprint(
    "Completeness",
    __name__,
    url_prefix="/completeness",
    description="Completness operations",
)


@blp.route("")
@blp.login_required
@blp.etag
@blp.arguments(CompletenessQueryArgsSchema, location="query")
@blp.response(200, CompletenessSchema)
def get_completeness(args):
    """Get timeseries data completeness

    Returns a report about data completeness, involving the number of values
    per time bucket and the sample interval.

    If the theoretical time interval for a timeseries is not defined in database,
    it is inferred from the maximum number of values per bucket.
    """
    try:
        timeseries = Timeseries.get_many_by_id(args["timeseries"])
    except TimeseriesNotFoundError as exc:
        abort(422, message=str(exc))
    if (data_state := TimeseriesDataState.get_by_id(args["data_state"])) is None:
        abort(422, errors={"query": {"data_state": "Unknown data state ID"}})

    completeness = compute_completeness(
        args["start_time"],
        args["end_time"],
        timeseries,
        data_state,
        args["bucket_width_value"],
        args["bucket_width_unit"],
        timezone=args["timezone"],
    )

    return completeness



"""Events API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import Event

from bemserver_api import AutoSchema, Schema, SortField


class EventSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = Event

    id = msa.auto_field(dump_only=True)
    timestamp = ma.fields.AwareDateTime()
    campaign_scope_id = ma.fields.Int(required=True)


class EventPutSchema(EventSchema):
    class Meta(EventSchema.Meta):
        exclude = ("campaign_scope_id", "timestamp")


class EventQueryArgsSchema(Schema):
    sort = SortField(("timestamp",))
    campaign_scope_id = ma.fields.Integer()
    source = ma.fields.Str()
    category_id = ma.fields.Int()
    level_id = ma.fields.Int()
    timestamp_min = ma.fields.AwareDateTime()
    timestamp_max = ma.fields.AwareDateTime()



"""Timeseries properties resources"""

from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import TimeseriesProperty

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import (
    TimeseriesPropertySchema,
    TimeseriesPropertyPutSchema,
    TimeseriesPropertyQueryArgsSchema,
)


blp = Blueprint(
    "TimeseriesProperty",
    __name__,
    url_prefix="/timeseries_properties",
    description="Operations on timeseries properties",
)


@blp.route("/")
class TimeseriesPropertiesViews(MethodView):
    @blp.login_required
    @blp.arguments(TimeseriesPropertyQueryArgsSchema, location="query")
    @blp.response(200, TimeseriesPropertySchema(many=True))
    def get(self, args):
        """List timeseries properties"""
        return TimeseriesProperty.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(TimeseriesPropertySchema)
    @blp.response(201, TimeseriesPropertySchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new timeseries property"""
        item = TimeseriesProperty.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class TimeseriesPropertyByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, TimeseriesPropertySchema)
    def get(self, item_id):
        """Get timeseries property by ID"""
        item = TimeseriesProperty.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(TimeseriesPropertyPutSchema)
    @blp.response(200, TimeseriesPropertySchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing timeseries property"""
        item = TimeseriesProperty.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, TimeseriesPropertySchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a timeseries property"""
        item = TimeseriesProperty.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, TimeseriesPropertySchema)
        item.delete()
        db.session.commit()



"""Events tests"""
import datetime as dt

import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

EVENTS_URL = "/events/"


class TestEventsApi:
    def test_events_api(
        self, app, users, campaign_scopes, event_categories, event_levels
    ):
        cs_1_id = campaign_scopes[0]
        cs_2_id = campaign_scopes[1]
        ec_1_id = event_categories[0]
        el_1_id = event_levels[0]
        el_2_id = event_levels[1]
        c1_st = dt.datetime(2020, 1, 1, tzinfo=dt.timezone.utc).isoformat()
        c2_et = dt.datetime(2021, 1, 1, tzinfo=dt.timezone.utc).isoformat()

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(EVENTS_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            event_1 = {
                "campaign_scope_id": cs_1_id,
                "timestamp": c1_st,
                "source": "Event source",
                "category_id": ec_1_id,
                "level_id": el_1_id,
            }
            ret = client.post(EVENTS_URL, json=event_1)
            assert ret.status_code == 201
            ret_val = ret.json
            event_1_id = ret_val.pop("id")
            event_1_etag = ret.headers["ETag"]

            # POST violating fkey constraint
            event = event_1.copy()
            event["level_id"] = DUMMY_ID
            ret = client.post(EVENTS_URL, json=event)
            assert ret.status_code == 409

            # GET list
            ret = client.get(EVENTS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == event_1_id

            # GET by id
            ret = client.get(f"{EVENTS_URL}{event_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == event_1_etag

            # PUT
            event_1["source"] = "Super event source"
            put_event = event_1.copy()
            del put_event["campaign_scope_id"]
            del put_event["timestamp"]
            ret = client.put(
                f"{EVENTS_URL}{event_1_id}",
                json=put_event,
                headers={"If-Match": event_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            event_1_etag = ret.headers["ETag"]
            assert ret_val == event_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{EVENTS_URL}{DUMMY_ID}",
                json=put_event,
                headers={"If-Match": event_1_etag},
            )
            assert ret.status_code == 404

            # POST
            event_2 = {
                "campaign_scope_id": cs_2_id,
                "timestamp": c2_et,
                "source": "Another event source",
                "category_id": ec_1_id,
                "level_id": el_2_id,
            }
            ret = client.post(EVENTS_URL, json=event_2)
            assert ret.status_code == 201
            ret_val = ret.json
            event_2_id = ret_val.pop("id")
            event_2_etag = ret.headers["ETag"]

            # GET list (filtered)
            ret = client.get(EVENTS_URL, query_string={"campaign_scope_id": cs_1_id})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == event_1_id
            ret = client.get(EVENTS_URL, query_string={"level_id": el_2_id})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == event_2_id
            ret = client.get(
                EVENTS_URL,
                query_string={
                    "campaign_scope_id": cs_2_id,
                    "level_id": el_1_id,
                },
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0
            ret = client.get(EVENTS_URL, query_string={"timestamp_min": c2_et})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            ret = client.get(EVENTS_URL, query_string={"timestamp_max": c1_st})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1

            # GET sorted list
            ret = client.get(EVENTS_URL, query_string={"sort": "timestamp"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2
            assert ret_val[0]["id"] == event_1_id
            ret = client.get(EVENTS_URL, query_string={"sort": "-timestamp"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2
            assert ret_val[0]["id"] == event_2_id

            # DELETE wrong ID -> 404
            # ETag is wrong but we get rejected before ETag check anyway
            ret = client.delete(
                f"{EVENTS_URL}{DUMMY_ID}",
                headers={"If-Match": "Dummy-ETag"},
            )
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(
                f"{EVENTS_URL}{event_1_id}",
                headers={"If-Match": event_1_etag},
            )
            assert ret.status_code == 204
            ret = client.delete(
                f"{EVENTS_URL}{event_2_id}",
                headers={"If-Match": event_2_etag},
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(EVENTS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{EVENTS_URL}{event_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    def test_events_as_user_api(
        self,
        app,
        users,
        campaign_scopes,
        campaigns,
        events,
        event_categories,
        event_levels,
    ):
        event_1_id = events[0]
        cs_1_id = campaign_scopes[0]
        ec_1_id = event_categories[0]
        el_1_id = event_levels[0]
        c1_st = dt.datetime(2021, 1, 1, tzinfo=dt.timezone.utc).isoformat()

        creds = users["Active"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(EVENTS_URL)
            assert ret.status_code == 200
            assert len(ret.json) == 1

            # POST
            event = {
                "campaign_scope_id": cs_1_id,
                "timestamp": c1_st,
                "source": "Event source",
                "category_id": ec_1_id,
                "level_id": el_1_id,
            }
            ret = client.post(EVENTS_URL, json=event)
            assert ret.status_code == 201

            # GET by id
            ret = client.get(f"{EVENTS_URL}{event_1_id}")
            assert ret.status_code == 200
            event_1_etag = ret.headers["ETag"]

            # DELETE
            ret = client.delete(
                f"{EVENTS_URL}{event_1_id}",
                headers={"If-Match": event_1_etag},
            )
            assert ret.status_code == 204

    def test_events_as_anonym_api(
        self, app, campaign_scopes, events, event_categories, event_levels
    ):
        event_1_id = events[0]
        cs_1_id = campaign_scopes[0]
        ec_1_id = event_categories[0]
        el_1_id = event_levels[0]
        c1_st = dt.datetime(2021, 1, 1, tzinfo=dt.timezone.utc).isoformat()

        client = app.test_client()

        # GET list
        ret = client.get(EVENTS_URL)
        assert ret.status_code == 401

        # POST
        event = {
            "campaign_scope_id": cs_1_id,
            "timestamp": c1_st,
            "source": "Event source",
            "category_id": ec_1_id,
            "level_id": el_1_id,
        }
        ret = client.post(EVENTS_URL, json=event)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{EVENTS_URL}{event_1_id}")
        assert ret.status_code == 401

        # DELETE
        # ETag is wrong but we get rejected before ETag check anyway
        ret = client.delete(
            f"{EVENTS_URL}{event_1_id}",
            headers={"If-Match": "Dummy-ETag"},
        )
        assert ret.status_code == 401



"""Timeseries by buildings API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import TimeseriesByBuilding

from bemserver_api import AutoSchema, Schema


class TimeseriesByBuildingSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = TimeseriesByBuilding

    id = msa.auto_field(dump_only=True)


class TimeseriesByBuildingQueryArgsSchema(Schema):
    building_id = ma.fields.Int()
    timeseries_id = ma.fields.Int()



"""Energy sources resources"""

from flask.views import MethodView

from bemserver_core.model import EnergySource

from bemserver_api import Blueprint

from .schemas import EnergySourceSchema


blp = Blueprint(
    "EnergySource",
    __name__,
    url_prefix="/energy_sources",
    description="Operations on energy sources",
)


@blp.route("/")
class EnergySourceViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, EnergySourceSchema(many=True))
    def get(self):
        """List energy sources"""
        return EnergySource.get()



"""Timeseries by events routes tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

TIMESERIES_BY_EVENTS_URL = "/timeseries_by_events/"
EVENTS_URL = "/events/"


class TestTimeseriesByEventApi:
    @pytest.mark.parametrize("timeseries", (4,), indirect=True)
    def test_timeseries_by_events_api(self, app, users, events, timeseries):

        creds = users["Chuck"]["creds"]
        event_1_id = events[0]
        event_2_id = events[1]
        ts_1_id = timeseries[0]
        ts_2_id = timeseries[1]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(TIMESERIES_BY_EVENTS_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            tbs_1 = {
                "event_id": event_1_id,
                "timeseries_id": ts_1_id,
            }
            ret = client.post(TIMESERIES_BY_EVENTS_URL, json=tbs_1)
            assert ret.status_code == 201
            ret_val = ret.json
            tbs_1_id = ret_val.pop("id")
            tbs_1_etag = ret.headers["ETag"]
            assert ret_val == tbs_1

            # POST violating unique constraint
            ret = client.post(TIMESERIES_BY_EVENTS_URL, json=tbs_1)
            assert ret.status_code == 409

            # POST event + TS from different campaign scopes
            tbs = {
                "event_id": event_2_id,
                "timeseries_id": ts_1_id,
            }
            ret = client.post(TIMESERIES_BY_EVENTS_URL, json=tbs)
            assert ret.status_code == 422
            ret_val = ret.json
            assert ret_val["errors"]["json"]["_schema"] == (
                "Event and timeseries must be in same campaign scope"
            )

            # GET list
            ret = client.get(TIMESERIES_BY_EVENTS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ts_1_id

            # GET by id
            ret = client.get(f"{TIMESERIES_BY_EVENTS_URL}{tbs_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == tbs_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == tbs_1

            # POST
            tbs_2 = {
                "event_id": event_2_id,
                "timeseries_id": ts_2_id,
            }
            ret = client.post(TIMESERIES_BY_EVENTS_URL, json=tbs_2)
            ret_val = ret.json
            tbs_2_id = ret_val.pop("id")

            # GET list
            ret = client.get(TIMESERIES_BY_EVENTS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(
                TIMESERIES_BY_EVENTS_URL,
                query_string={"event_id": event_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == tbs_1_id

            # DELETE wrong ID -> 404
            ret = client.delete(f"{TIMESERIES_BY_EVENTS_URL}{DUMMY_ID}")
            assert ret.status_code == 404

            # DELETE event cascade
            ret = client.get(f"{EVENTS_URL}{event_1_id}")
            event_1_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{EVENTS_URL}{event_1_id}", headers={"If-Match": event_1_etag}
            )
            assert ret.status_code == 204

            # DELETE
            ret = client.delete(f"{TIMESERIES_BY_EVENTS_URL}{tbs_1_id}")
            assert ret.status_code == 404
            ret = client.delete(f"{TIMESERIES_BY_EVENTS_URL}{tbs_2_id}")
            assert ret.status_code == 204

            # GET list
            ret = client.get(TIMESERIES_BY_EVENTS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{TIMESERIES_BY_EVENTS_URL}{tbs_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    @pytest.mark.parametrize("timeseries", (4,), indirect=True)
    def test_timeseries_by_events_as_user_api(
        self, app, users, events, timeseries, timeseries_by_events
    ):

        user_creds = users["Active"]["creds"]
        event_1_id = events[0]
        event_2_id = events[1]
        ts_2_id = timeseries[1]
        ts_3_id = timeseries[2]
        tbs_1_id = timeseries_by_events[0]
        tbs_2_id = timeseries_by_events[1]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(TIMESERIES_BY_EVENTS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            tbs_1 = ret_val[0]
            assert tbs_1.pop("id") == tbs_1_id

            # POST
            tbs_3 = {
                "event_id": event_1_id,
                "timeseries_id": ts_3_id,
            }
            ret = client.post(TIMESERIES_BY_EVENTS_URL, json=tbs_3)
            assert ret.status_code == 201
            ret_val = ret.json

            # POST not in campaign scope
            tbs = {
                "event_id": event_2_id,
                "timeseries_id": ts_2_id,
            }
            ret = client.post(TIMESERIES_BY_EVENTS_URL, json=tbs)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{TIMESERIES_BY_EVENTS_URL}{tbs_1_id}")
            assert ret.status_code == 200
            tbs_1_etag = ret.headers["ETag"]

            # GET by id not in campaign scope
            ret = client.get(f"{TIMESERIES_BY_EVENTS_URL}{tbs_2_id}")
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{TIMESERIES_BY_EVENTS_URL}{tbs_1_id}",
                headers={"If-Match": tbs_1_etag},
            )
            assert ret.status_code == 204

    def test_timeseries_by_events_as_anonym_api(
        self, app, events, timeseries, timeseries_by_events
    ):
        event_1_id = events[0]
        ts_2_id = timeseries[1]
        tbs_1_id = timeseries_by_events[0]

        client = app.test_client()

        # GET list
        ret = client.get(TIMESERIES_BY_EVENTS_URL)
        assert ret.status_code == 401

        # POST
        tbs_3 = {
            "event_id": event_1_id,
            "timeseries_id": ts_2_id,
        }
        ret = client.post(TIMESERIES_BY_EVENTS_URL, json=tbs_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{TIMESERIES_BY_EVENTS_URL}{tbs_1_id}")
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(f"{TIMESERIES_BY_EVENTS_URL}{tbs_1_id}")
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Users routes tests"""
from tests.common import AuthHeader


DUMMY_ID = "69"

USERS_URL = "/users/"


class TestUsersApi:
    def test_users_api(self, app, users):

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(USERS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            nb_init_users = len(ret_val)
            init_admin_users_ids = {u["id"] for u in ret_val if u["is_admin"]}
            init_inactive_users_ids = {u["id"] for u in ret_val if not u["is_active"]}

            # POST
            user_1 = {
                "name": "User 1",
                "email": "user_1@test.com",
                "password": "p@ss",
            }
            user_1_expected_get = {
                "name": "User 1",
                "email": "user_1@test.com",
                "is_admin": False,
                "is_active": True,
            }
            ret = client.post(USERS_URL, json=user_1)
            assert ret.status_code == 201
            ret_val = ret.json
            user_1_id = ret_val.pop("id")
            user_1_etag = ret.headers["ETag"]
            assert ret_val == user_1_expected_get

            # POST violating unique constraint
            ret = client.post(USERS_URL, json=user_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(USERS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == nb_init_users + 1

            # GET by id
            ret = client.get(f"{USERS_URL}{user_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == user_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == user_1_expected_get

            # PUT
            user_1["email"] = "user_1-spam@test.com"
            user_1_expected_get["email"] = "user_1-spam@test.com"
            ret = client.put(
                f"{USERS_URL}{user_1_id}",
                json=user_1,
                headers={"If-Match": user_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            user_1_etag = ret.headers["ETag"]
            assert ret_val == user_1_expected_get

            # PUT wrong ID -> 404
            ret = client.put(
                f"{USERS_URL}{DUMMY_ID}", json=user_1, headers={"If-Match": user_1_etag}
            )
            assert ret.status_code == 404

            # Set admin
            ret = client.put(
                f"{USERS_URL}{user_1_id}/set_admin",
                json={"value": True},
                headers={"If-Match": user_1_etag},
            )
            assert ret.status_code == 204
            user_1_etag = ret.headers["ETag"]

            # Set admin wrong ID -> 404
            ret = client.put(
                f"{USERS_URL}{DUMMY_ID}/set_admin",
                json={"value": True},
                headers={"If-Match": user_1_etag},
            )
            assert ret.status_code == 404

            # POST user 2
            user_2 = {
                "name": "User 2",
                "email": "user_2@test.com",
                "password": "pa55",
            }
            ret = client.post(USERS_URL, json=user_2)
            assert ret.status_code == 201
            ret_val = ret.json
            user_2_id = ret_val.pop("id")
            user_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            user_2["email"] = user_1["email"]
            ret = client.put(
                f"{USERS_URL}{user_2_id}",
                json=user_2,
                headers={"If-Match": user_2_etag},
            )
            assert ret.status_code == 409

            # Set inactive
            ret = client.put(
                f"{USERS_URL}{user_2_id}/set_active",
                json={"value": False},
                headers={"If-Match": user_2_etag},
            )
            assert ret.status_code == 204
            user_2_etag = ret.headers["ETag"]

            # Set inactive wrong ID -> 404
            ret = client.put(
                f"{USERS_URL}{DUMMY_ID}/set_active",
                json={"value": False},
                headers={"If-Match": user_2_etag},
            )
            assert ret.status_code == 404

            # GET list
            ret = client.get(USERS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == nb_init_users + 2

            # GET list (filtered)
            ret = client.get(USERS_URL, query_string={"is_admin": True})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == len(init_admin_users_ids) + 1
            assert {u["id"] for u in ret_val} == init_admin_users_ids | {user_1_id}
            ret = client.get(USERS_URL, query_string={"is_active": False})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == len(init_inactive_users_ids) + 1
            assert {u["id"] for u in ret_val} == init_inactive_users_ids | {user_2_id}

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{USERS_URL}{DUMMY_ID}", headers={"If-Match": user_1_etag}
            )
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(
                f"{USERS_URL}{user_1_id}", headers={"If-Match": user_1_etag}
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(USERS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == nb_init_users + 1

            # GET by id -> 404
            ret = client.get(f"{USERS_URL}{user_1_id}")
            assert ret.status_code == 404

    def test_users_as_user_api(self, app, users):

        creds = users["Active"]["creds"]
        user_1_id = users["Active"]["id"]
        user_2_id = users["Inactive"]["id"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(USERS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1

            # POST
            new_user = {
                "name": "New user",
                "email": "new_user@test.com",
                "password": "p@ss",
            }
            ret = client.post(USERS_URL, json=new_user)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{USERS_URL}{user_1_id}")
            assert ret.status_code == 200
            user_1 = ret.json
            user_1_etag = ret.headers["ETag"]
            ret = client.get(f"{USERS_URL}{user_2_id}")
            assert ret.status_code == 403

            # PUT
            user_1 = {
                "name": "Superactive",
                "password": "@ctive",
                "email": "active@test.com",
            }
            ret = client.put(
                f"{USERS_URL}{user_1_id}",
                json=user_1,
                headers={"If-Match": user_1_etag},
            )
            assert ret.status_code == 200
            user_1_etag = ret.headers["ETag"]
            ret = client.put(
                f"{USERS_URL}{user_2_id}",
                json=new_user,
                headers={"If-Match": user_1_etag},
            )
            # ETag is wrong but we get rejected before ETag check anyway
            assert ret.status_code == 403

            # Set admin
            ret = client.put(
                f"{USERS_URL}{user_1_id}/set_admin",
                json={"value": True},
                headers={"If-Match": user_1_etag},
            )
            assert ret.status_code == 403

            # Set inactive
            ret = client.put(
                f"{USERS_URL}{user_1_id}/set_active",
                json={"value": False},
                headers={"If-Match": user_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{USERS_URL}{user_1_id}", headers={"If-Match": user_1_etag}
            )
            assert ret.status_code == 403

    def test_users_as_anonym_api(self, app, users):

        user_1_id = users["Active"]["id"]

        client = app.test_client()

        # GET list
        ret = client.get(USERS_URL)
        assert ret.status_code == 401

        # POST
        new_user = {
            "name": "New user",
            "email": "new_user@test.com",
            "password": "p@ss",
        }
        ret = client.post(USERS_URL, json=new_user)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{USERS_URL}{user_1_id}")
        assert ret.status_code == 401

        # PUT
        user_1 = {
            "name": "Superactive",
            "password": "@ctive",
            "email": "active@test.com",
        }
        ret = client.put(
            f"{USERS_URL}{user_1_id}", json=user_1, headers={"If-Match": "Dummy-ETag"}
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # Set admin
        ret = client.put(
            f"{USERS_URL}{user_1_id}/set_admin",
            json={"value": True},
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # Set inactive
        ret = client.put(
            f"{USERS_URL}{user_1_id}/set_active",
            json={"value": False},
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{USERS_URL}{user_1_id}", headers={"If-Match": "Dummy-ETag"}
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""About resources"""
import bemserver_core
import bemserver_api
from bemserver_api import Blueprint

from .schemas import AboutSchema


blp = Blueprint(
    "About", __name__, url_prefix="/about", description="Informations about BEMServer"
)


@blp.get("/")
@blp.etag
@blp.response(200, AboutSchema)
def about():
    """Provide information about BEMServer

    Returns bemserver_core and bemserver_api versions.
    """
    return {
        "versions": {
            "bemserver_core": bemserver_core.__version__,
            "bemserver_api": bemserver_api.__version__,
        }
    }



"""Timeseries by zones API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import TimeseriesByZone

from bemserver_api import AutoSchema, Schema


class TimeseriesByZoneSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = TimeseriesByZone

    id = msa.auto_field(dump_only=True)


class TimeseriesByZoneQueryArgsSchema(Schema):
    zone_id = ma.fields.Int()
    timeseries_id = ma.fields.Int()



"""User by user groups routes tests"""
from tests.common import AuthHeader


DUMMY_ID = "69"

USERS_URL = "/users/"
USER_GROUPS_URL = "/user_groups/"
USERS_BY_USER_GROUPS_URL = "/users_by_user_groups/"


class TestUsersByCampaignsApi:
    def test_users_by_user_groups_api(self, app, users, user_groups):

        user_1_id = users["Active"]["id"]
        user_2_id = users["Inactive"]["id"]
        ug_1_id = user_groups[0]
        ug_2_id = user_groups[1]

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(USERS_BY_USER_GROUPS_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            ubug_1 = {"user_group_id": ug_1_id, "user_id": user_1_id}
            ret = client.post(USERS_BY_USER_GROUPS_URL, json=ubug_1)
            assert ret.status_code == 201
            ret_val = ret.json
            ubug_1_id = ret_val.pop("id")
            ubug_1_etag = ret.headers["ETag"]

            # POST violating unique constraint
            ret = client.post(USERS_BY_USER_GROUPS_URL, json=ubug_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(USERS_BY_USER_GROUPS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ubug_1_id

            # GET by id
            ret = client.get(f"{USERS_BY_USER_GROUPS_URL}{ubug_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == ubug_1_etag

            # POST
            ubug_2 = {"user_group_id": ug_2_id, "user_id": user_2_id}
            ret = client.post(USERS_BY_USER_GROUPS_URL, json=ubug_2)
            assert ret.status_code == 201
            ret_val = ret.json
            ubug_2_id = ret_val.pop("id")

            # GET list (filtered)
            ret = client.get(
                USERS_BY_USER_GROUPS_URL, query_string={"user_id": user_1_id}
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ubug_1_id
            assert ret_val[0]["user_id"] == user_1_id
            assert ret_val[0]["user_group_id"] == ug_1_id
            ret = client.get(
                USERS_BY_USER_GROUPS_URL, query_string={"user_group_id": ug_2_id}
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ubug_2_id
            assert ret_val[0]["user_id"] == user_2_id
            assert ret_val[0]["user_group_id"] == ug_2_id
            ret = client.get(
                USERS_BY_USER_GROUPS_URL,
                query_string={"user_id": user_1_id, "user_group_id": ug_2_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # DELETE wrong ID -> 404
            ret = client.delete(f"{USERS_BY_USER_GROUPS_URL}{DUMMY_ID}")
            assert ret.status_code == 404

            # DELETE
            ubug_3 = {"user_group_id": ug_1_id, "user_id": user_2_id}
            ret = client.post(USERS_BY_USER_GROUPS_URL, json=ubug_3)
            ret_val = ret.json
            ubug_3_id = ret_val.pop("id")
            assert ret.status_code == 201
            ret = client.delete(f"{USERS_BY_USER_GROUPS_URL}{ubug_3_id}")
            assert ret.status_code == 204

            # DELETE user cascade
            ret = client.get(f"{USERS_URL}{user_1_id}")
            user_1_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{USERS_URL}{user_1_id}", headers={"If-Match": user_1_etag}
            )
            assert ret.status_code == 204

            # DELETE user_group cascade
            ret = client.get(f"{USER_GROUPS_URL}{ug_2_id}")
            ug_2_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{USER_GROUPS_URL}{ug_2_id}", headers={"If-Match": ug_2_etag}
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(USERS_BY_USER_GROUPS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{USERS_BY_USER_GROUPS_URL}{ubug_1_id}")
            assert ret.status_code == 404

    def test_users_by_user_groups_as_user_api(
        self, app, users, user_groups, users_by_user_groups
    ):

        user_creds = users["Active"]["creds"]
        user_1_id = users["Active"]["id"]
        user_2_id = users["Inactive"]["id"]
        ug_1_id = user_groups[0]
        ug_2_id = user_groups[1]
        ubug_1_id = users_by_user_groups[0]
        ubug_2_id = users_by_user_groups[1]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(USERS_BY_USER_GROUPS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            ubug_1 = ret_val[0]
            assert ubug_1["id"] == ubug_1_id

            # POST
            ubug_3 = {"user_group_id": ug_1_id, "user_id": user_1_id}
            ret = client.post(USERS_BY_USER_GROUPS_URL, json=ubug_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{USERS_BY_USER_GROUPS_URL}{ubug_1_id}")
            assert ret.status_code == 200
            ret = client.get(f"{USERS_BY_USER_GROUPS_URL}{ubug_2_id}")
            assert ret.status_code == 403

            # GET list (filtered)
            ret = client.get(
                USERS_BY_USER_GROUPS_URL, query_string={"user_id": user_1_id}
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == ubug_1_id
            assert ret_val[0]["user_id"] == user_1_id
            assert ret_val[0]["user_group_id"] == ug_1_id
            ret = client.get(
                USERS_BY_USER_GROUPS_URL, query_string={"user_group_id": ug_2_id}
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0
            ret = client.get(
                USERS_BY_USER_GROUPS_URL, query_string={"user_id": user_2_id}
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # DELETE
            ret = client.delete(f"{USERS_BY_USER_GROUPS_URL}{ubug_1_id}")
            assert ret.status_code == 403

    def test_users_by_user_groups_as_anonym_api(
        self, app, users, user_groups, users_by_user_groups
    ):

        user_1_id = users["Active"]["id"]
        ug_1_id = user_groups[0]
        ubug_1_id = users_by_user_groups[0]

        client = app.test_client()

        # GET list
        ret = client.get(USERS_BY_USER_GROUPS_URL)
        assert ret.status_code == 401

        # POST
        ubug_1 = {"user_group_id": ug_1_id, "user_id": user_1_id}
        ret = client.post(USERS_BY_USER_GROUPS_URL, json=ubug_1)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{USERS_BY_USER_GROUPS_URL}{ubug_1_id}")
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(f"{USERS_BY_USER_GROUPS_URL}{ubug_1_id}")
        assert ret.status_code == 401



"""Energy consumption API schemas"""
import marshmallow as ma

from bemserver_api import Schema
from bemserver_api.extensions.ma_fields import Timezone
from bemserver_api.resources.timeseries_data.schemas import TimeseriesBucketWidthSchema


class EnergyConsumptionSchema(Schema):
    timestamps = ma.fields.List(
        ma.fields.AwareDateTime,
        metadata={
            "description": "Time index (value is bucket start time)",
        },
    )
    energy = ma.fields.Dict(
        keys=ma.fields.String(metadata={"description": "Energy"}),
        values=ma.fields.Dict(
            keys=ma.fields.String(metadata={"description": "End use"}),
            values=ma.fields.List(
                ma.fields.Float(metadata={"description": "Value (Wh)"})
            ),
        ),
    )


class EnergyConsumptionQueryArgsSchema(TimeseriesBucketWidthSchema):
    start_time = ma.fields.AwareDateTime(
        required=True,
        metadata={
            "description": "Initial datetime",
        },
    )
    end_time = ma.fields.AwareDateTime(
        required=True,
        metadata={
            "description": "End datetime (excluded from the interval)",
        },
    )
    timezone = Timezone(
        load_default="UTC",
        metadata={
            "description": "Timezone to use for the aggreagation",
        },
    )



"""User groups routes tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

USER_GROUPS_URL = "/user_groups/"


class TestUserGroupsApi:
    def test_user_groups_api(self, app, users):

        creds = users["Chuck"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(USER_GROUPS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # POST
            ug_1 = {
                "name": "User group 1",
            }
            ret = client.post(USER_GROUPS_URL, json=ug_1)
            assert ret.status_code == 201
            ret_val = ret.json
            ug_1_id = ret_val.pop("id")
            ug_1_etag = ret.headers["ETag"]
            assert ret_val == ug_1

            # POST violating unique constraint
            ret = client.post(USER_GROUPS_URL, json=ug_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(USER_GROUPS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1

            # GET by id
            ret = client.get(f"{USER_GROUPS_URL}{ug_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == ug_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == ug_1

            # PUT
            ug_1["name"] = "Super user group"
            ret = client.put(
                f"{USER_GROUPS_URL}{ug_1_id}",
                json=ug_1,
                headers={"If-Match": ug_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            ug_1_etag = ret.headers["ETag"]
            assert ret_val == ug_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{USER_GROUPS_URL}{DUMMY_ID}",
                json=ug_1,
                headers={"If-Match": ug_1_etag},
            )
            assert ret.status_code == 404

            # POST user 2
            ug_2 = {
                "name": "User group 2",
            }
            ret = client.post(USER_GROUPS_URL, json=ug_2)
            assert ret.status_code == 201
            ret_val = ret.json
            ug_2_id = ret_val.pop("id")
            ug_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            ug_2["name"] = ug_1["name"]
            ret = client.put(
                f"{USER_GROUPS_URL}{ug_2_id}",
                json=ug_2,
                headers={"If-Match": ug_2_etag},
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(USER_GROUPS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{USER_GROUPS_URL}{DUMMY_ID}", headers={"If-Match": ug_1_etag}
            )
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(
                f"{USER_GROUPS_URL}{ug_1_id}", headers={"If-Match": ug_1_etag}
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(USER_GROUPS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1

            # GET by id -> 404
            ret = client.get(f"{USER_GROUPS_URL}{ug_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    def test_user_groups_as_user_api(self, app, users, user_groups):

        creds = users["Active"]["creds"]
        ug_1_id = user_groups[0]
        ug_2_id = user_groups[1]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(USER_GROUPS_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1

            # POST
            new_ug = {
                "name": "New user group",
            }
            ret = client.post(USER_GROUPS_URL, json=new_ug)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{USER_GROUPS_URL}{ug_1_id}")
            assert ret.status_code == 200
            ug_1 = ret.json
            ug_1_etag = ret.headers["ETag"]
            ret = client.get(f"{USER_GROUPS_URL}{ug_2_id}")
            assert ret.status_code == 403

            # PUT
            ug_1 = {
                "name": "Super user group 1",
            }
            ret = client.put(
                f"{USER_GROUPS_URL}{ug_1_id}",
                json=ug_1,
                headers={"If-Match": ug_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{USER_GROUPS_URL}{ug_1_id}", headers={"If-Match": ug_1_etag}
            )
            assert ret.status_code == 403

    def test_user_groups_as_anonym_api(self, app, users):

        ug_1_id = users["Active"]["id"]

        client = app.test_client()

        # GET list
        ret = client.get(USER_GROUPS_URL)
        assert ret.status_code == 401

        # POST
        new_ug = {
            "name": "New user group",
        }
        ret = client.post(USER_GROUPS_URL, json=new_ug)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{USER_GROUPS_URL}{ug_1_id}")
        assert ret.status_code == 401

        # PUT
        ug_1 = {
            "name": "Super user group",
        }
        ret = client.put(
            f"{USER_GROUPS_URL}{ug_1_id}", json=ug_1, headers={"If-Match": "Dummy-ETag"}
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{USER_GROUPS_URL}{ug_1_id}", headers={"If-Match": "Dummy-ETag"}
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Event levels API schemas"""

import marshmallow_sqlalchemy as msa

from bemserver_core.model import EventLevel

from bemserver_api import AutoSchema


class EventLevelSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = EventLevel

    id = msa.auto_field(dump_only=True)



"""Energy sources API schemas"""

import marshmallow_sqlalchemy as msa

from bemserver_core.model import EnergySource

from bemserver_api import AutoSchema


class EnergySourceSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = EnergySource

    id = msa.auto_field(dump_only=True)



"""Default configuration"""


class Config:
    """Default configuration"""

    # SQLAlchemy parameters
    SQLALCHEMY_DATABASE_URI = ""
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    # API parameters
    API_TITLE = "BEMServer API"
    OPENAPI_JSON_PATH = "api-spec.json"
    OPENAPI_URL_PREFIX = "/"
    OPENAPI_RAPIDOC_PATH = "/"
    OPENAPI_RAPIDOC_URL = "https://cdn.jsdelivr.net/npm/rapidoc/dist/rapidoc-min.js"
    OPENAPI_RAPIDOC_CONFIG = {
        "theme": "dark",
        "show-header": "false",
        "render-style": "focused",
    }



"""Timeseries tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

TIMESERIES_URL = "/timeseries/"
CAMPAIGNS_URL = "/campaigns/"
CAMPAIGN_SCOPES_URL = "/campaign_scopes/"


class TestTimeseriesApi:
    def test_timeseries_api(self, app, users, campaigns, campaign_scopes):

        creds = users["Chuck"]["creds"]
        campaign_1_id = campaigns[0]
        campaign_2_id = campaigns[1]
        cs_1_id = campaign_scopes[0]
        cs_2_id = campaign_scopes[1]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(TIMESERIES_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            timeseries_1 = {
                "name": "Timeseries 1",
                "description": "Timeseries example 1",
                "campaign_id": campaign_1_id,
                "campaign_scope_id": cs_1_id,
                "unit_symbol": "°C",
            }
            ret = client.post(TIMESERIES_URL, json=timeseries_1)
            assert ret.status_code == 201
            ret_val = ret.json
            timeseries_1_id = ret_val.pop("id")
            timeseries_1_etag = ret.headers["ETag"]
            assert ret_val == timeseries_1

            # POST violating unique constraint
            ret = client.post(TIMESERIES_URL, json=timeseries_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(TIMESERIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == timeseries_1_id

            # GET by id
            ret = client.get(f"{TIMESERIES_URL}{timeseries_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == timeseries_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == timeseries_1

            # PUT
            del timeseries_1["description"]
            timeseries_1_put = timeseries_1.copy()
            del timeseries_1_put["campaign_id"]
            del timeseries_1_put["campaign_scope_id"]
            ret = client.put(
                f"{TIMESERIES_URL}{timeseries_1_id}",
                json=timeseries_1_put,
                headers={"If-Match": timeseries_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            timeseries_1_etag = ret.headers["ETag"]
            assert ret_val == timeseries_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{TIMESERIES_URL}{DUMMY_ID}",
                json=timeseries_1_put,
                headers={"If-Match": timeseries_1_etag},
            )
            assert ret.status_code == 404

            # POST TS 2
            timeseries_2 = {
                "name": "Timeseries 2",
                "campaign_id": campaign_1_id,
                "campaign_scope_id": cs_2_id,
            }
            ret = client.post(TIMESERIES_URL, json=timeseries_2)
            ret_val = ret.json

            # PUT violating unique constraint
            timeseries_1_put["name"] = timeseries_2["name"]
            ret = client.put(
                f"{TIMESERIES_URL}{timeseries_1_id}",
                json=timeseries_1_put,
                headers={"If-Match": timeseries_1_etag},
            )
            assert ret.status_code == 409

            # GET list with filters
            ret = client.get(TIMESERIES_URL, query_string={"name": "Timeseries 1"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == timeseries_1_id
            ret = client.get(TIMESERIES_URL, query_string={"unit_symbol": "°C"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == timeseries_1_id
            ret = client.get(TIMESERIES_URL, query_string={"unit_symbol": "kWh"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # DELETE
            timeseries_3 = {
                "name": "Timeseries 3",
                "description": "Timeseries example 3",
                "campaign_id": campaign_2_id,
                "campaign_scope_id": cs_1_id,
            }
            ret = client.post(TIMESERIES_URL, json=timeseries_3)
            ret_val = ret.json
            assert ret.status_code == 201
            timeseries_3_id = ret_val.pop("id")
            timeseries_3_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{TIMESERIES_URL}{timeseries_3_id}",
                headers={"If-Match": timeseries_3_etag},
            )
            assert ret.status_code == 204

            # DELETE campaign scope cascade
            ret = client.get(f"{CAMPAIGN_SCOPES_URL}{cs_1_id}")
            cs_1_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{CAMPAIGN_SCOPES_URL}{cs_1_id}", headers={"If-Match": cs_1_etag}
            )
            assert ret.status_code == 204

            # DELETE campaign cascade
            ret = client.get(f"{CAMPAIGNS_URL}{campaign_2_id}")
            campaign_2_etag = ret.headers["ETag"]
            ret = client.delete(
                f"{CAMPAIGNS_URL}{campaign_2_id}", headers={"If-Match": campaign_2_etag}
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(TIMESERIES_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # GET by id -> 404
            ret = client.get(f"{TIMESERIES_URL}{timeseries_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    def test_timeseries_as_user_api(
        self,
        app,
        users,
        timeseries,
        campaigns,
        campaign_scopes,
        sites,
        buildings,
        storeys,
        spaces,
        zones,
        timeseries_by_sites,
        timeseries_by_buildings,
        timeseries_by_storeys,
        timeseries_by_spaces,
        timeseries_by_zones,
    ):

        campaign_1_id = campaigns[0]
        cs_1_id = campaign_scopes[0]
        timeseries_1_id = timeseries[0]
        timeseries_2_id = timeseries[1]

        creds = users["Active"]["creds"]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(TIMESERIES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["campaign_id"] == campaign_1_id
            assert ret_val[0]["campaign_scope_id"] == cs_1_id
            assert ret_val[0]["id"] == timeseries_1_id

            # GET list for a structural element
            for struct_elmt_type in ["site", "building", "storey", "space", "zone"]:
                ret = client.get(
                    TIMESERIES_URL,
                    query_string={f"{struct_elmt_type}_id": 666},
                )
                assert ret.status_code == 200
                ret_val = ret.json
                assert len(ret_val) == 0
                ret = client.get(
                    TIMESERIES_URL,
                    query_string={f"{struct_elmt_type}_id": 1},
                )
                assert ret.status_code == 200
                ret_val = ret.json
                assert len(ret_val) == 1
                assert ret_val[0]["campaign_id"] == campaign_1_id
                assert ret_val[0]["campaign_scope_id"] == cs_1_id
                assert ret_val[0]["id"] == timeseries_1_id

            # GET list using "in_name"
            ret = client.get(TIMESERIES_URL, query_string={"in_name": "Toto"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0
            ret = client.get(TIMESERIES_URL, query_string={"in_name": "series"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["campaign_id"] == campaign_1_id
            assert ret_val[0]["campaign_scope_id"] == cs_1_id
            assert ret_val[0]["id"] == timeseries_1_id

            # POST
            timeseries_1 = {
                "name": "Timeseries 1",
                "description": "Timeseries example 1",
                "campaign_id": campaign_1_id,
                "campaign_scope_id": cs_1_id,
            }
            ret = client.post(TIMESERIES_URL, json=timeseries_1)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{TIMESERIES_URL}{timeseries_1_id}")
            assert ret.status_code == 200
            ts_1_etag = ret.headers["ETag"]

            # GET by id, user not in group
            ret = client.get(f"{TIMESERIES_URL}{timeseries_2_id}")
            assert ret.status_code == 403

            # PUT
            del timeseries_1["description"]
            del timeseries_1["campaign_id"]
            del timeseries_1["campaign_scope_id"]
            ret = client.put(
                f"{TIMESERIES_URL}{timeseries_1_id}",
                json=timeseries_1,
                headers={"If-Match": ts_1_etag},
            )
            # ETag is wrong but we get rejected before ETag check anyway
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{TIMESERIES_URL}{timeseries_1_id}",
                headers={"If-Match": ts_1_etag},
            )
            # ETag is wrong but we get rejected before ETag check anyway
            assert ret.status_code == 403

    def test_timeseries_as_anonym_api(
        self, app, users, timeseries, campaigns, campaign_scopes
    ):

        campaign_1_id = campaigns[0]
        cs_1_id = campaign_scopes[0]
        timeseries_1_id = timeseries[0]

        client = app.test_client()

        # GET list
        ret = client.get(TIMESERIES_URL)
        assert ret.status_code == 401

        # POST
        timeseries_1 = {
            "name": "Timeseries 1",
            "description": "Timeseries example 1",
            "campaign_id": campaign_1_id,
            "campaign_scope_id": cs_1_id,
        }
        ret = client.post(TIMESERIES_URL, json=timeseries_1)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{TIMESERIES_URL}{timeseries_1_id}")
        assert ret.status_code == 401

        # PUT
        del timeseries_1["description"]
        del timeseries_1["campaign_id"]
        del timeseries_1["campaign_scope_id"]
        ret = client.put(
            f"{TIMESERIES_URL}{timeseries_1_id}",
            json=timeseries_1,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{TIMESERIES_URL}{timeseries_1_id}",
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Energy consumption resources"""
from flask_smorest import abort

from bemserver_core.process.energy_consumption import (
    compute_energy_consumption_breakdown_for_site,
    compute_energy_consumption_breakdown_for_building,
)
from bemserver_core.model import Site, Building


from bemserver_api import Blueprint

from .schemas import EnergyConsumptionSchema, EnergyConsumptionQueryArgsSchema


blp = Blueprint(
    "Energy consumption",
    __name__,
    url_prefix="/energy_consumption",
    description="Energy consumption operations",
)


@blp.route("/site/<int:site_id>")
@blp.login_required
@blp.etag
@blp.arguments(EnergyConsumptionQueryArgsSchema, location="query")
@blp.response(200, EnergyConsumptionSchema)
def get_energy_consumption_breakdown_for_site(args, site_id):
    """Get energy consumption breakdown for a site"""
    site = Site.get_by_id(site_id)
    if site is None:
        abort(404)

    brkdwn = compute_energy_consumption_breakdown_for_site(
        site,
        args["start_time"],
        args["end_time"],
        args["bucket_width_value"],
        args["bucket_width_unit"],
        timezone=args["timezone"],
    )

    return brkdwn


@blp.route("/building/<int:building_id>")
@blp.login_required
@blp.etag
@blp.arguments(EnergyConsumptionQueryArgsSchema, location="query")
@blp.response(200, EnergyConsumptionSchema)
def get_energy_consumption_breakdown_for_building(args, building_id):
    """Get energy consumption breakdown for a building"""
    building = Building.get_by_id(building_id)
    if building is None:
        abort(404)

    brkdwn = compute_energy_consumption_breakdown_for_building(
        building,
        args["start_time"],
        args["end_time"],
        args["bucket_width_value"],
        args["bucket_width_unit"],
        timezone=args["timezone"],
    )

    return brkdwn



"""Database access"""
from bemserver_core.database import db


def init_app(app):
    """Init DB accessor with app

    Sets DB engine using app config.
    Adds app contextteardown method to close DB session.
    """
    db.set_db_url(app.config["SQLALCHEMY_DATABASE_URI"])

    @app.teardown_appcontext
    def cleanup(_):
        db.session.remove()



"""Timeseries by sites resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import TimeseriesBySite

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import (
    TimeseriesBySiteSchema,
    TimeseriesBySiteQueryArgsSchema,
)


blp = Blueprint(
    "TimeseriesBySite",
    __name__,
    url_prefix="/timeseries_by_sites",
    description="Operations on timeseries x site associations",
)


@blp.route("/")
class TimeseriesBySiteViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(TimeseriesBySiteQueryArgsSchema, location="query")
    @blp.response(200, TimeseriesBySiteSchema(many=True))
    def get(self, args):
        """List timeseries x site associations"""
        return TimeseriesBySite.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(TimeseriesBySiteSchema)
    @blp.response(201, TimeseriesBySiteSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new timeseries x site association"""
        item = TimeseriesBySite.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class TimeseriesBySiteByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, TimeseriesBySiteSchema)
    def get(self, item_id):
        """Get timeseries x site association ID"""
        item = TimeseriesBySite.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(TimeseriesBySiteSchema)
    @blp.response(200, TimeseriesBySiteSchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing timeseries x site association"""
        item = TimeseriesBySite.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, TimeseriesBySiteSchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a timeseries x site association"""
        item = TimeseriesBySite.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, TimeseriesBySiteSchema)
        item.delete()
        db.session.commit()



"""About routes tests"""
import contextlib

import bemserver_core
import bemserver_api

import pytest

from tests.common import AuthHeader


ABOUT_URL = "/about/"


class TestAboutApi:
    @pytest.mark.parametrize("user", ("user", "admin", "anonym"))
    def test_about_api(self, app, user, users):

        if user == "user":
            auth_ctx = AuthHeader(users["Active"]["creds"])
        elif user == "admin":
            auth_ctx = AuthHeader(users["Chuck"]["creds"])
        else:
            auth_ctx = contextlib.nullcontext()

        client = app.test_client()

        with auth_ctx:

            # GET about info
            ret = client.get(ABOUT_URL)
            assert ret.status_code == 200
            assert ret.json == {
                "versions": {
                    "bemserver_core": bemserver_core.__version__,
                    "bemserver_api": bemserver_api.__version__,
                }
            }



#!/usr/bin/env python3
"""BEMServer API"""

from setuptools import setup, find_packages


# Get the long description from the README file
with open("README.rst", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="bemserver-api",
    version="0.4.0",
    description="BEMServer API",
    long_description=long_description,
    url="https://github.com/BEMServer/bemserver-api",
    author="Nobatek/INEF4",
    author_email="jlafrechoux@nobatek.inef4.com",
    license="AGPLv3+",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        (
            "License :: OSI Approved :: "
            "GNU Affero General Public License v3 or later (AGPLv3+)"
        ),
    ],
    python_requires=">=3.9",
    install_requires=[
        "flask>=2.0.0",
        "python-dotenv>=0.9.0",
        "marshmallow>=3.18.0,<4.0",
        "marshmallow-sqlalchemy>=0.24.0",
        "flask_smorest>=0.40.0,<0.41",
        "flask-httpauth>=0.5.0",
        "bemserver-core>=0.3,<0.4",
    ],
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
)



"""Zone property data routes tests"""
import copy
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

ZONE_PROPERTY_DATA_URL = "/zone_property_data/"


class TestZonePropertyDataApi:
    def test_zone_property_data_api(self, app, users, zones, zone_properties):

        creds = users["Chuck"]["creds"]
        zone_1_id = zones[0]
        zone_2_id = zones[1]
        zone_p_1_id = zone_properties[0]
        zone_p_2_id = zone_properties[1]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(ZONE_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            zpd_1 = {
                "zone_id": zone_1_id,
                "zone_property_id": zone_p_1_id,
                "value": "12",
            }
            ret = client.post(ZONE_PROPERTY_DATA_URL, json=zpd_1)
            assert ret.status_code == 201
            ret_val = ret.json
            zpd_1_id = ret_val.pop("id")
            zpd_1_etag = ret.headers["ETag"]
            assert ret_val == zpd_1

            # POST violating unique constraint
            ret = client.post(ZONE_PROPERTY_DATA_URL, json=zpd_1)
            assert ret.status_code == 409

            # POST wrong value type
            zpd_post = {
                "zone_id": zone_1_id,
                "zone_property_id": zone_p_2_id,
                "value": "wrong type",
            }
            ret = client.post(ZONE_PROPERTY_DATA_URL, json=zpd_post)
            assert ret.status_code == 422

            # GET list
            ret = client.get(ZONE_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == zone_p_1_id

            # GET by id
            ret = client.get(f"{ZONE_PROPERTY_DATA_URL}{zpd_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == zpd_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == zpd_1

            # PUT
            zpd_1["value"] = "69"
            ret = client.put(
                f"{ZONE_PROPERTY_DATA_URL}{zpd_1_id}",
                json=zpd_1,
                headers={"If-Match": zpd_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            zpd_1_etag = ret.headers["ETag"]
            assert ret_val == zpd_1

            # PUT wrong value type
            zpd_1_put = copy.deepcopy(zpd_1)
            zpd_1_put["value"] = "wrong type"
            ret = client.put(
                f"{ZONE_PROPERTY_DATA_URL}{zpd_1_id}",
                json=zpd_1_put,
                headers={"If-Match": zpd_1_etag},
            )
            assert ret.status_code == 422

            # PUT wrong ID -> 404
            ret = client.put(
                f"{ZONE_PROPERTY_DATA_URL}{DUMMY_ID}",
                json=zpd_1,
                headers={"If-Match": zpd_1_etag},
            )
            assert ret.status_code == 404

            # POST sep 2
            zpd_2 = {
                "zone_id": zone_2_id,
                "zone_property_id": zone_p_2_id,
                "value": "42",
            }
            ret = client.post(ZONE_PROPERTY_DATA_URL, json=zpd_2)
            ret_val = ret.json
            zpd_2_id = ret_val.pop("id")
            zpd_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            zpd_1["zone_id"] = zpd_2["zone_id"]
            zpd_1["zone_property_id"] = zpd_2["zone_property_id"]
            ret = client.put(
                f"{ZONE_PROPERTY_DATA_URL}{zpd_1_id}",
                json=zpd_1,
                headers={"If-Match": zpd_1_etag},
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(ZONE_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(
                ZONE_PROPERTY_DATA_URL,
                query_string={"zone_id": zone_1_id},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == zpd_1_id

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{ZONE_PROPERTY_DATA_URL}{DUMMY_ID}",
                headers={"If-Match": zpd_1_etag},
            )
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(
                f"{ZONE_PROPERTY_DATA_URL}{zpd_1_id}",
                headers={"If-Match": zpd_1_etag},
            )
            assert ret.status_code == 204
            ret = client.delete(
                f"{ZONE_PROPERTY_DATA_URL}{zpd_2_id}",
                headers={"If-Match": zpd_2_etag},
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(ZONE_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{ZONE_PROPERTY_DATA_URL}{zpd_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    def test_zone_property_data_as_user_api(
        self, app, users, zones, zone_properties, zone_property_data
    ):

        user_creds = users["Active"]["creds"]
        zone_1_id = zones[0]
        zone_p_1_id = zone_properties[0]
        zone_p_1_id = zone_properties[0]
        zpd_1_id = zone_property_data[0]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(ZONE_PROPERTY_DATA_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            zpd_1 = ret_val[0]
            assert zpd_1.pop("id") == zpd_1_id

            # POST
            zpd_3 = {
                "zone_id": zone_1_id,
                "zone_property_id": zone_p_1_id,
                "value": "12",
            }
            ret = client.post(ZONE_PROPERTY_DATA_URL, json=zpd_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{ZONE_PROPERTY_DATA_URL}{zpd_1_id}")
            assert ret.status_code == 200
            zpd_1_etag = ret.headers["ETag"]

            # PUT
            zpd_1["value"] = "69"
            ret = client.put(
                f"{ZONE_PROPERTY_DATA_URL}{zpd_1_id}",
                json=zpd_1,
                headers={"If-Match": zpd_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{ZONE_PROPERTY_DATA_URL}{zpd_1_id}",
                headers={"If-Match": zpd_1_etag},
            )
            assert ret.status_code == 403

    def test_zone_property_data_as_anonym_api(
        self, app, zones, zone_properties, zone_property_data
    ):
        zone_1_id = zones[0]
        zone_p_1_id = zone_properties[0]
        zone_p_2_id = zone_properties[1]
        zpd_1_id = zone_property_data[0]

        client = app.test_client()

        # GET list
        ret = client.get(ZONE_PROPERTY_DATA_URL)
        assert ret.status_code == 401

        # POST
        zpd_3 = {
            "zone_id": zone_1_id,
            "zone_property_id": zone_p_2_id,
            "value": "12",
        }
        ret = client.post(ZONE_PROPERTY_DATA_URL, json=zpd_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{ZONE_PROPERTY_DATA_URL}{zpd_1_id}")
        assert ret.status_code == 401

        # PUT
        zpd_1 = {
            "zone_id": zone_1_id,
            "zone_property_id": zone_p_1_id,
            "value": "12",
        }
        ret = client.put(
            f"{ZONE_PROPERTY_DATA_URL}{zpd_1_id}",
            json=zpd_1,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{ZONE_PROPERTY_DATA_URL}{zpd_1_id}",
            headers={"If-Match": "Dummy-Etag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""User groups API schemas"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import UserGroup

from bemserver_api import AutoSchema, Schema, SortField


class UserGroupSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = UserGroup

    id = msa.auto_field(dump_only=True)
    name = msa.auto_field(validate=ma.validate.Length(1, 80))


class UserGroupQueryArgsSchema(Schema):
    sort = SortField(("name",))
    name = ma.fields.Str()



"""Building property data API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import BuildingPropertyData

from bemserver_api import AutoSchema, Schema


class BuildingPropertyDataSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = BuildingPropertyData

    id = msa.auto_field(dump_only=True)


class BuildingPropertyDataQueryArgsSchema(Schema):
    building_id = ma.fields.Int()
    building_property_id = ma.fields.Int()



"""Analysis resources"""
from bemserver_api import Blueprint

from .completeness.routes import blp as completeness_blp
from .energy_consumption.routes import blp as energy_consumption_blp


blp = Blueprint(
    "Analysis", __name__, url_prefix="/analysis", description="Data analysis operations"
)


blp.register_blueprint(completeness_blp)
blp.register_blueprint(energy_consumption_blp)


def register_blueprints(api):
    api.register_blueprint(blp)



"""Zone resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import Zone

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import ZoneSchema, ZonePutSchema, ZoneQueryArgsSchema


blp = Blueprint(
    "Zone", __name__, url_prefix="/zones", description="Operations on zones"
)


@blp.route("/")
class ZoneViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(ZoneQueryArgsSchema, location="query")
    @blp.response(200, ZoneSchema(many=True))
    def get(self, args):
        """List zones"""
        return Zone.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(ZoneSchema)
    @blp.response(201, ZoneSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new zone"""
        item = Zone.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class ZoneByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, ZoneSchema)
    def get(self, item_id):
        """Get zone by ID"""
        item = Zone.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(ZonePutSchema)
    @blp.response(200, ZoneSchema)
    @blp.catch_integrity_error
    def put(self, new_item, item_id):
        """Update an existing zone"""
        item = Zone.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, ZoneSchema)
        item.update(**new_item)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a zone"""
        item = Zone.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, ZoneSchema)
        item.delete()
        db.session.commit()



"""ST_CheckMissingByCampaign resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.scheduled_tasks import ST_CheckMissingByCampaign

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import (
    ST_CheckMissingByCampaignSchema,
    ST_CheckMissingByCampaignQueryArgsSchema,
    ST_CheckMissingByCampaignPutSchema,
    ST_CheckMissingByCampaignFullSchema,
    ST_CheckMissingByCampaignFullQueryArgsSchema,
)


blp = Blueprint(
    "ST_CheckMissingByCampaign",
    __name__,
    url_prefix="/st_check_missings_by_campaigns",
    description="Operations on check missings scheduled task x campaign associations",
)


@blp.route("/")
class ST_CheckMissingByCampaignViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(ST_CheckMissingByCampaignQueryArgsSchema, location="query")
    @blp.response(200, ST_CheckMissingByCampaignSchema(many=True))
    def get(self, args):
        """List check missings scheduled tasks x campaign associations"""
        return ST_CheckMissingByCampaign.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(ST_CheckMissingByCampaignSchema)
    @blp.response(201, ST_CheckMissingByCampaignSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new check missings scheduled tasks x campaign association"""
        item = ST_CheckMissingByCampaign.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class ST_CheckMissingByCampaignByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, ST_CheckMissingByCampaignSchema)
    def get(self, item_id):
        """Get check missings scheduled tasks x campaign association by ID"""
        item = ST_CheckMissingByCampaign.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.etag
    @blp.arguments(ST_CheckMissingByCampaignPutSchema)
    @blp.response(200, ST_CheckMissingByCampaignSchema)
    def put(self, item_data, item_id):
        """Update check missings scheduled tasks x campaign association by ID"""
        item = ST_CheckMissingByCampaign.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, ST_CheckMissingByCampaignSchema)
        item.update(**item_data)
        db.session.commit()
        return item

    @blp.login_required
    @blp.etag
    @blp.response(204)
    def delete(self, item_id):
        """Delete a check missings scheduled tasks x campaign associations"""
        item = ST_CheckMissingByCampaign.get_by_id(item_id)
        if item is None:
            abort(404)
        blp.check_etag(item, ST_CheckMissingByCampaignSchema)
        item.delete()
        db.session.commit()


@blp.route("/full")
@blp.login_required
@blp.etag
@blp.arguments(ST_CheckMissingByCampaignFullQueryArgsSchema, location="query")
@blp.response(200, ST_CheckMissingByCampaignFullSchema(many=True))
def get_full(args):
    """List check missings service state for all campaigns"""
    return ST_CheckMissingByCampaign.get_all(**args)



"""Sites API schemas"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import Site

from bemserver_api import AutoSchema, Schema, SortField


class SiteSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = Site

    id = msa.auto_field(dump_only=True)
    name = msa.auto_field(validate=ma.validate.Length(1, 80))
    campaign_id = ma.fields.Int(required=True)


class SitePutSchema(SiteSchema):
    class Meta(SiteSchema.Meta):
        exclude = ("campaign_id",)


class SiteQueryArgsSchema(Schema):
    sort = SortField(("name",))
    name = ma.fields.Str()
    campaign_id = ma.fields.Int()
    ifc_id = ma.fields.String()



"""Input/Output API schemas"""
import marshmallow as ma
from flask_smorest.fields import Upload

from bemserver_api import Schema


class SitesCSVUploadQueryArgsSchema(Schema):
    campaign_id = ma.fields.Int(required=True)


class SitesCSVUploadFileSchema(Schema):
    sites_csv = Upload()
    buildings_csv = Upload()
    storeys_csv = Upload()
    spaces_csv = Upload()
    zones_csv = Upload()


class TimeseriesCSVUploadQueryArgsSchema(Schema):
    campaign_id = ma.fields.Int(required=True)


class TimeseriesCSVUploadFileSchema(Schema):
    timeseries_csv = Upload(required=True)



"""Zone properties API schemas"""

import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import ZoneProperty

from bemserver_api import AutoSchema, Schema
from ..structural_element_properties.schemas import StructuralElementPropertySchema


class ZonePropertySchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = ZoneProperty

    id = msa.auto_field(dump_only=True)
    structural_element_property = ma.fields.Nested(
        StructuralElementPropertySchema(exclude=("id",)),
        dump_only=True,
    )


class ZonePropertyQueryArgsSchema(Schema):
    structural_element_property_id = ma.fields.Int()



"""Energy consumption timeseries by buildings API schemas"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import EnergyConsumptionTimeseriesByBuilding

from bemserver_api import AutoSchema, Schema


class EnergyConsumptionTimeseriesByBuildingSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = EnergyConsumptionTimeseriesByBuilding

    id = msa.auto_field(dump_only=True)
    wh_conversion_factor = msa.auto_field(metadata={"default": 1})


class EnergyConsumptionTimeseriesByBuildingQueryArgsSchema(Schema):
    campaign_id = ma.fields.Int()
    building_id = ma.fields.Int()
    source_id = ma.fields.Int()
    end_use_id = ma.fields.Int()
    timeseries_id = ma.fields.Int()



"""ST_CheckMissingByCampaign API schemas"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.scheduled_tasks import ST_CheckMissingByCampaign

from bemserver_api import AutoSchema, Schema, SortField


class ST_CheckMissingByCampaignSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = ST_CheckMissingByCampaign

    id = msa.auto_field(dump_only=True)
    is_enabled = msa.auto_field(metadata={"default": True})


class ST_CheckMissingByCampaignPutSchema(Schema):
    is_enabled = ma.fields.Bool()


class ST_CheckMissingByCampaignFullSchema(Schema):
    id = ma.fields.Int()
    campaign_id = ma.fields.Int()
    campaign_name = ma.fields.Str()
    is_enabled = ma.fields.Bool()


class ST_CheckMissingByCampaignQueryArgsSchema(Schema):
    campaign_id = ma.fields.Int()


class ST_CheckMissingByCampaignFullQueryArgsSchema(Schema):
    sort = SortField(("campaign_name",))
    is_enabled = ma.fields.Bool()
    campaign_id = ma.fields.Int()
    in_campaign_name = ma.fields.Str(
        metadata={
            "description": (
                "Search for items whose name contains this input value"
                " (case insensitive)"
            ),
        }
    )



"""Users by user groups resources"""
from flask.views import MethodView
from flask_smorest import abort

from bemserver_core.model import UserByUserGroup

from bemserver_api import Blueprint
from bemserver_api.database import db

from .schemas import UserByUserGroupSchema, UserByUserGroupQueryArgsSchema


blp = Blueprint(
    "UserByUserGroup",
    __name__,
    url_prefix="/users_by_user_groups",
    description="Operations on users x user groups associations",
)


@blp.route("/")
class UserByUserGroupViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.arguments(UserByUserGroupQueryArgsSchema, location="query")
    @blp.response(200, UserByUserGroupSchema(many=True))
    def get(self, args):
        """List user x user group associations"""
        return UserByUserGroup.get(**args)

    @blp.login_required
    @blp.etag
    @blp.arguments(UserByUserGroupSchema)
    @blp.response(201, UserByUserGroupSchema)
    @blp.catch_integrity_error
    def post(self, new_item):
        """Add a new user x user group association"""
        item = UserByUserGroup.new(**new_item)
        db.session.commit()
        return item


@blp.route("/<int:item_id>")
class UserByUserGroupByIdViews(MethodView):
    @blp.login_required
    @blp.etag
    @blp.response(200, UserByUserGroupSchema)
    def get(self, item_id):
        """Get user x user group association by ID"""
        item = UserByUserGroup.get_by_id(item_id)
        if item is None:
            abort(404)
        return item

    @blp.login_required
    @blp.response(204)
    def delete(self, item_id):
        """Delete a user x user group association"""
        item = UserByUserGroup.get_by_id(item_id)
        if item is None:
            abort(404)
        item.delete()
        db.session.commit()



"""Test authentication extension"""
from flask import jsonify

from bemserver_api import Blueprint

from bemserver_core.authorization import get_current_user


class TestAuthentication:
    def test_auth_login_required(self, app, users):

        active_user_creds = users["Active"]["creds"]
        active_user_invalid_creds = users["Active"]["invalid_creds"]
        inactive_user_creds = users["Inactive"]["creds"]
        api = app.extensions["flask-smorest"]["ext_obj"]
        blp = Blueprint("AuthTest", __name__, url_prefix="/auth_test")

        @blp.route("/auth")
        @blp.login_required
        @blp.response(200)
        def auth():
            return get_current_user().name

        @blp.route("/no_auth")
        @blp.response(204)
        def no_auth():
            return None

        api.register_blueprint(blp)
        client = app.test_client()

        # Anonymous user
        headers = {}
        resp = client.get("/auth_test/auth", headers=headers)
        assert resp.status_code == 401
        resp = client.get("/auth_test/no_auth", headers=headers)
        assert resp.status_code == 204

        # Inactive user
        headers = {"Authorization": "Basic " + inactive_user_creds}
        resp = client.get("/auth_test/auth", headers=headers)
        assert resp.status_code == 403
        resp = client.get("/auth_test/no_auth", headers=headers)
        assert resp.status_code == 204

        # Active user with invalid creds (bad password)
        headers = {"Authorization": "Basic " + active_user_invalid_creds}
        resp = client.get("/auth_test/auth", headers=headers)
        assert resp.status_code == 401
        resp = client.get("/auth_test/no_auth", headers=headers)
        assert resp.status_code == 204

        # Active user
        headers = {"Authorization": "Basic " + active_user_creds}
        resp = client.get("/auth_test/auth", headers=headers)
        assert resp.status_code == 200
        resp = client.get("/auth_test/no_auth", headers=headers)
        assert resp.status_code == 204

        # Check OpenAPI spec
        spec = api.spec.to_dict()
        assert spec["components"]["securitySchemes"]["BasicAuthentication"] == {
            "type": "http",
            "scheme": "basic",
        }
        auth_spec = spec["paths"]["/auth_test/auth"]
        assert auth_spec["get"]["responses"]["401"] == {
            "$ref": "#/components/responses/UNAUTHORIZED"
        }
        assert auth_spec["get"]["security"] == [{"BasicAuthentication": []}]
        no_auth_spec = spec["paths"]["/auth_test/no_auth"]
        assert "401" not in no_auth_spec["get"]["responses"]
        assert "security" not in no_auth_spec["get"]

    def test_auth_current_user(self, app, users):

        active_user_creds = users["Active"]["creds"]
        api = app.extensions["flask-smorest"]["ext_obj"]

        blp = Blueprint("AuthTest", __name__, url_prefix="/auth_test")

        @blp.route("/user")
        @blp.login_required
        @blp.response(200)
        def user():
            return jsonify(blp.current_user().name)

        api.register_blueprint(blp)
        client = app.test_client()

        # Active user
        headers = {"Authorization": "Basic " + active_user_creds}
        resp = client.get("/auth_test/user", headers=headers)
        assert resp.json == "Active"



"""Users API schemas"""
import marshmallow as ma
import marshmallow_sqlalchemy as msa

from bemserver_core.model import User

from bemserver_api import AutoSchema, Schema, SortField


class UserSchema(AutoSchema):
    class Meta(AutoSchema.Meta):
        model = User
        exclude = ("_is_active", "_is_admin")

    id = msa.auto_field(dump_only=True)
    name = msa.auto_field(validate=ma.validate.Length(1, 80))
    email = ma.fields.Email(required=True)
    password = msa.auto_field(validate=ma.validate.Length(1, 80), load_only=True)
    is_admin = ma.fields.Boolean(dump_only=True)
    is_active = ma.fields.Boolean(dump_only=True)


class UserQueryArgsSchema(Schema):
    sort = SortField(("name", "email", "is_admin", "is_active"))
    name = ma.fields.Str()
    email = ma.fields.Str()
    is_admin = ma.fields.Boolean()
    is_active = ma.fields.Boolean()


class BooleanValueSchema(Schema):
    value = ma.fields.Boolean()



"""Zones routes tests"""
import pytest

from tests.common import AuthHeader


DUMMY_ID = "69"

ZONES_URL = "/zones/"


class TestZonesApi:
    def test_zones_api(self, app, users, campaigns):

        creds = users["Chuck"]["creds"]
        campaign_1_id = campaigns[0]

        client = app.test_client()

        with AuthHeader(creds):

            # GET list
            ret = client.get(ZONES_URL)
            assert ret.status_code == 200
            assert ret.json == []

            # POST
            zone_1 = {
                "name": "Zone 1",
                "campaign_id": campaign_1_id,
            }
            ret = client.post(ZONES_URL, json=zone_1)
            assert ret.status_code == 201
            ret_val = ret.json
            zone_1_id = ret_val.pop("id")
            zone_1_etag = ret.headers["ETag"]
            assert ret_val == zone_1

            # POST violating unique constraint
            ret = client.post(ZONES_URL, json=zone_1)
            assert ret.status_code == 409

            # GET list
            ret = client.get(ZONES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == zone_1_id

            # GET by id
            ret = client.get(f"{ZONES_URL}{zone_1_id}")
            assert ret.status_code == 200
            assert ret.headers["ETag"] == zone_1_etag
            ret_val = ret.json
            ret_val.pop("id")
            assert ret_val == zone_1

            # PUT
            zone_1["description"] = "Fantastic zone"
            zone_1_put = zone_1.copy()
            del zone_1_put["campaign_id"]
            ret = client.put(
                f"{ZONES_URL}{zone_1_id}",
                json=zone_1_put,
                headers={"If-Match": zone_1_etag},
            )
            assert ret.status_code == 200
            ret_val = ret.json
            ret_val.pop("id")
            zone_1_etag = ret.headers["ETag"]
            assert ret_val == zone_1

            # PUT wrong ID -> 404
            ret = client.put(
                f"{ZONES_URL}{DUMMY_ID}",
                json=zone_1_put,
                headers={"If-Match": zone_1_etag},
            )
            assert ret.status_code == 404

            # POST zone 2
            zone_2 = {
                "name": "Zone 2",
                "campaign_id": campaign_1_id,
            }
            ret = client.post(ZONES_URL, json=zone_2)
            ret_val = ret.json
            zone_2_id = ret_val.pop("id")
            zone_2_etag = ret.headers["ETag"]

            # PUT violating unique constraint
            zone_1_put["name"] = zone_2["name"]
            ret = client.put(
                f"{ZONES_URL}{zone_1_id}",
                json=zone_1_put,
                headers={"If-Match": zone_1_etag},
            )
            assert ret.status_code == 409

            # GET list
            ret = client.get(ZONES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 2

            # GET list with filters
            ret = client.get(ZONES_URL, query_string={"name": "Zone 1"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == zone_1_id

            # DELETE wrong ID -> 404
            ret = client.delete(
                f"{ZONES_URL}{DUMMY_ID}", headers={"If-Match": zone_1_etag}
            )
            assert ret.status_code == 404

            # DELETE
            ret = client.delete(
                f"{ZONES_URL}{zone_1_id}", headers={"If-Match": zone_1_etag}
            )
            assert ret.status_code == 204
            ret = client.delete(
                f"{ZONES_URL}{zone_2_id}", headers={"If-Match": zone_2_etag}
            )
            assert ret.status_code == 204

            # GET list
            ret = client.get(ZONES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 0

            # GET by id -> 404
            ret = client.get(f"{ZONES_URL}{zone_1_id}")
            assert ret.status_code == 404

    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaigns")
    def test_zones_as_user_api(self, app, users, campaigns, zones):

        user_creds = users["Active"]["creds"]
        campaign_2_id = campaigns[1]
        zone_1_id = zones[0]
        zone_2_id = zones[1]

        client = app.test_client()

        with AuthHeader(user_creds):

            # GET list
            ret = client.get(ZONES_URL)
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            zone_1 = ret_val[0]
            assert zone_1.pop("id") == zone_1_id

            # GET list with filters
            ret = client.get(ZONES_URL, query_string={"name": "Zone 1"})
            assert ret.status_code == 200
            ret_val = ret.json
            assert len(ret_val) == 1
            assert ret_val[0]["id"] == zone_1_id
            ret = client.get(ZONES_URL, query_string={"name": "Zone 2"})
            assert ret.status_code == 200
            assert not ret.json

            # POST
            zone_3 = {
                "name": "Zone 3",
                "campaign_id": campaign_2_id,
            }
            ret = client.post(ZONES_URL, json=zone_3)
            assert ret.status_code == 403

            # GET by id
            ret = client.get(f"{ZONES_URL}{zone_1_id}")
            assert ret.status_code == 200
            zone_1_etag = ret.headers["ETag"]

            ret = client.get(f"{ZONES_URL}{zone_2_id}")
            assert ret.status_code == 403

            # PUT
            zone_1["description"] = "Fantastic zone"
            zone_1_put = zone_1.copy()
            del zone_1_put["campaign_id"]
            ret = client.put(
                f"{ZONES_URL}{zone_1_id}",
                json=zone_1_put,
                headers={"If-Match": zone_1_etag},
            )
            assert ret.status_code == 403

            # DELETE
            ret = client.delete(
                f"{ZONES_URL}{zone_1_id}", headers={"If-Match": zone_1_etag}
            )
            assert ret.status_code == 403

    def test_zones_as_anonym_api(self, app, zones, campaigns):

        zone_1_id = zones[0]
        campaign_1_id = campaigns[0]

        client = app.test_client()

        # GET list
        ret = client.get(ZONES_URL)
        assert ret.status_code == 401

        # POST
        zone_3 = {
            "name": "Zone 3",
            "campaign_id": campaign_1_id,
        }
        ret = client.post(ZONES_URL, json=zone_3)
        assert ret.status_code == 401

        # GET by id
        ret = client.get(f"{ZONES_URL}{zone_1_id}")
        assert ret.status_code == 401

        # PUT
        zone_1 = {
            "name": "Super Zone 1",
        }
        ret = client.put(
            f"{ZONES_URL}{zone_1_id}",
            json=zone_1,
            headers={"If-Match": "Dummy-ETag"},
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401

        # DELETE
        ret = client.delete(
            f"{ZONES_URL}{zone_1_id}", headers={"If-Match": "Dummy-Etag"}
        )
        # ETag is wrong but we get rejected before ETag check anyway
        assert ret.status_code == 401



"""Input / Output tests"""
import contextlib
import io

import pytest

from tests.common import AuthHeader

INPUT_OUTPUT_URL = "/io/"
INPUT_OUTPUT_SITES_URL = f"{INPUT_OUTPUT_URL}sites"
INPUT_OUTPUT_TIMESERIES_URL = f"{INPUT_OUTPUT_URL}timeseries"
BUILDINGS_URL = "/buildings/"
TIMESERIES_URL = "/timeseries/"

DUMMY_ID = "69"


class TestInputOutputSites:
    @pytest.mark.parametrize("user", ("admin", "user", "anonym"))
    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    @pytest.mark.usefixtures("site_properties")
    @pytest.mark.usefixtures("building_properties")
    @pytest.mark.usefixtures("storey_properties")
    @pytest.mark.usefixtures("space_properties")
    @pytest.mark.usefixtures("zone_properties")
    def test_sites_csv_post(self, app, user, users, campaigns):

        campaign_1_id = campaigns[0]

        if user == "admin":
            creds = users["Chuck"]["creds"]
            auth_context = AuthHeader(creds)
        elif user == "user":
            creds = users["Active"]["creds"]
            auth_context = AuthHeader(creds)
        else:
            auth_context = contextlib.nullcontext()

        client = app.test_client()

        sites_csv = (
            "Name,Description,IFC_ID,Area\n"
            "Site 1,Great site 1,abcdefghijklmnopqrtsuv,1000\n"
            "Site 2,Great site 2,,2000\n"
        )
        buildings_csv = (
            "Name,Description,Site,IFC_ID,Area\n"
            "Building 1,Great building 1,Site 1,bcdefghijklmnopqrtsuvw,1000\n"
            "Building 2,Great building 2,Site 2,,2000\n"
        )
        storeys_csv = (
            "Name,Description,Site,Building,IFC_ID,Area\n"
            "Storey 1,Great storey 1,Site 1,Building 1,cdefghijklmnopqrtsuvwx,1000\n"
            "Storey 2,Great storey 2,Site 2,Building 2,,2000\n"
        )
        spaces_csv = (
            "Name,Description,Site,Building,Storey,IFC_ID,Area\n"
            "Storey 1,Great storey 1,Site 1,Building 1,Storey 1,"
            "defghijklmnopqrtsuvwxy,1000\n"
            "Storey 2,Great storey 2,Site 2,Building 2,Storey 2,"
            ",2000\n"
        )
        zones_csv = (
            "Name,Description,IFC_ID,Area\n"
            "Zone 1,Great zone 1,efghijklmnopqrtsuvwxyz,1000\n"
            "Zone 2,Great zone 2,,2000\n"
        )

        with auth_context:

            ret = client.post(
                INPUT_OUTPUT_SITES_URL,
                query_string={
                    "campaign_id": campaign_1_id,
                },
                data={
                    "sites_csv": (io.BytesIO(sites_csv.encode()), "sites.csv"),
                    "buildings_csv": (
                        io.BytesIO(buildings_csv.encode()),
                        "buildings.csv",
                    ),
                    "storeys_csv": (io.BytesIO(storeys_csv.encode()), "storeys.csv"),
                    "spaces_csv": (io.BytesIO(spaces_csv.encode()), "spaces.csv"),
                    "zones_csv": (io.BytesIO(zones_csv.encode()), "zones.csv"),
                },
            )
            if user == "anonym":
                assert ret.status_code == 401
            elif user == "user":
                assert ret.status_code == 403
            else:
                assert ret.status_code == 201
                ret = client.get(BUILDINGS_URL)
                assert ret.status_code == 200
                assert len(ret.json) == 2

    @pytest.mark.usefixtures("site_properties")
    @pytest.mark.usefixtures("building_properties")
    def test_sites_csv_post_sites_buildings(self, app, users, campaigns):

        campaign_1_id = campaigns[0]

        creds = users["Chuck"]["creds"]
        auth_context = AuthHeader(creds)

        client = app.test_client()

        sites_csv = (
            "Name,Description,IFC_ID,Area\n"
            "Site 1,Great site 1,,1000\n"
            "Site 2,Great site 2,,2000\n"
        )
        buildings_csv = (
            "Name,Description,Site,IFC_ID,Area\n"
            "Building 1,Great building 1,Site 1,,1000\n"
            "Building 2,Great building 2,Site 2,,2000\n"
        )

        with auth_context:

            ret = client.post(
                INPUT_OUTPUT_SITES_URL,
                query_string={
                    "campaign_id": campaign_1_id,
                },
                data={
                    "sites_csv": (io.BytesIO(sites_csv.encode()), "sites.csv"),
                    "buildings_csv": (
                        io.BytesIO(buildings_csv.encode()),
                        "buildings.csv",
                    ),
                },
            )
            assert ret.status_code == 201
            ret = client.get(BUILDINGS_URL)
            assert ret.status_code == 200
            assert len(ret.json) == 2

    @pytest.mark.parametrize(
        "buildings_csv",
        (
            # Missing column in header
            "Name,Description,Site",
            # Missing column in row
            "Name,Description,Site,IFC_ID\nBuilding 1,,",
            # Duplicate building
            "Name,Description,Site,IFC_ID\nBuilding 3,,Site 1,\nBuilding 3,,Site 1,",
            # Unknown site
            "Name,Description,Site,IFC_ID\nBuilding 1,,Dummy",
        ),
    )
    @pytest.mark.usefixtures("sites")
    def test_sites_csv_post_errors(self, app, users, campaigns, buildings_csv):

        campaign_1_id = campaigns[0]

        creds = users["Chuck"]["creds"]
        auth_context = AuthHeader(creds)

        client = app.test_client()

        with auth_context:

            ret = client.post(
                INPUT_OUTPUT_SITES_URL,
                query_string={
                    "campaign_id": campaign_1_id,
                },
                data={
                    "buildings_csv": (
                        io.BytesIO(buildings_csv.encode()),
                        "buildings.csv",
                    ),
                },
            )
            assert ret.status_code == 422

    def test_sites_csv_post_unknown_campaign(self, app, users):

        creds = users["Chuck"]["creds"]
        auth_context = AuthHeader(creds)

        client = app.test_client()

        sites_csv = (
            "Name,Description,IFC_ID\n"
            "Site 1,Great site 1,\n"
            "Site 1,Great site 2,\n"
        )

        with auth_context:

            ret = client.post(
                INPUT_OUTPUT_SITES_URL,
                query_string={
                    "campaign_id": DUMMY_ID,
                },
                data={
                    "sites_csv": (io.BytesIO(sites_csv.encode()), "sites.csv"),
                },
            )
            assert ret.status_code == 422
            ret_val = ret.json
            assert "campaign_id" in ret_val["errors"]["query"]


class TestInputOutputTimeseries:
    @pytest.mark.parametrize("user", ("admin", "user", "anonym"))
    @pytest.mark.usefixtures("users_by_user_groups")
    @pytest.mark.usefixtures("user_groups_by_campaign_scopes")
    @pytest.mark.usefixtures("sites")
    @pytest.mark.usefixtures("buildings")
    @pytest.mark.usefixtures("storeys")
    @pytest.mark.usefixtures("spaces")
    @pytest.mark.usefixtures("zones")
    def test_timeseries_csv_post(self, app, user, users, campaigns):

        campaign_1_id = campaigns[0]

        if user == "admin":
            creds = users["Chuck"]["creds"]
            auth_context = AuthHeader(creds)
        elif user == "user":
            creds = users["Active"]["creds"]
            auth_context = AuthHeader(creds)
        else:
            auth_context = contextlib.nullcontext()

        client = app.test_client()

        timeseries_csv = (
            "Name,Description,Unit,Campaign scope,Site,Building,"
            "Storey,Space,Zone,Min,Max\n"
            "Space_1_Temp,Temperature,°C,Campaign 1 - Scope 1,Site 1,Building 1,"
            "Storey 1,Space 1,Zone 1,-10,60\n"
            "Space_2_Temp,Temperature,°C,Campaign 1 - Scope 1,Site 1,Building 1,"
            "Storey 1,Space 1,Zone 1,-10,60\n"
        )

        with auth_context:

            ret = client.post(
                INPUT_OUTPUT_TIMESERIES_URL,
                query_string={
                    "campaign_id": campaign_1_id,
                },
                data={
                    "timeseries_csv": (
                        io.BytesIO(timeseries_csv.encode()),
                        "timeseries.csv",
                    ),
                },
            )
            if user == "anonym":
                assert ret.status_code == 401
            elif user == "user":
                assert ret.status_code == 403
            else:
                assert ret.status_code == 201
                ret = client.get(TIMESERIES_URL)
                assert ret.status_code == 200
                assert len(ret.json) == 2

    @pytest.mark.parametrize(
        "timeseries_csv",
        (
            # Missing column in header
            (
                "Name,Description,Unit,Campaign scope,\n"
                "Space_1_Temp,Temperature,°C,Campaign 1 - Scope 1,,,,,\n"
            ),
            # Duplicate timeseries
            (
                "Name,Description,Unit,Campaign scope,Site,Building,Storey,Space,Zone\n"
                "Space_1_Temp,Temperature,°C,Campaign 1 - Scope 1,,,,,\n"
                "Space_1_Temp,Temperature,°C,Campaign 1 - Scope 1,,,,,\n"
            ),
            # Unknown site
            (
                "Name,Description,Unit,Campaign scope,Site,Building,Storey,Space,Zone\n"
                "Space_1_Temp,Temperature,°C,Campaign 1 - Scope 1,Dummy,,,,\n"
            ),
        ),
    )
    @pytest.mark.usefixtures("campaign_scopes")
    def test_timeseries_csv_post_errors(self, app, users, campaigns, timeseries_csv):

        campaign_1_id = campaigns[0]

        creds = users["Chuck"]["creds"]
        auth_context = AuthHeader(creds)

        client = app.test_client()

        with auth_context:

            ret = client.post(
                INPUT_OUTPUT_TIMESERIES_URL,
                query_string={
                    "campaign_id": campaign_1_id,
                },
                data={
                    "timeseries_csv": (
                        io.BytesIO(timeseries_csv.encode()),
                        "timeseries.csv",
                    ),
                },
            )
            assert ret.status_code == 422

    def test_timeseries_csv_post_unknown_campaign(self, app, users):

        creds = users["Chuck"]["creds"]
        auth_context = AuthHeader(creds)

        client = app.test_client()

        timeseries_csv = (
            "Name,Description,Unit,Campaign scope,Site,Building,"
            "Storey,Space,Zone\n"
            "Space_1_Temp,Temperature,°C,Campaign 1 - Scope 1,Site 1,Building 1,"
            "Storey 1,Space 1,Zone 1\n"
            "Space_2_Temp,Temperature,°C,Campaign 1 - Scope 1,Site 1,Building 1,"
            "Storey 1,Space 1,Zone 1\n"
        )

        with auth_context:

            ret = client.post(
                INPUT_OUTPUT_TIMESERIES_URL,
                query_string={
                    "campaign_id": DUMMY_ID,
                },
                data={
                    "timeseries_csv": (
                        io.BytesIO(timeseries_csv.encode()),
                        "timeseries.csv",
                    ),
                },
            )
            assert ret.status_code == 422
            ret_val = ret.json
            assert "campaign_id" in ret_val["errors"]["query"]



"""BEMServer API extensions"""
from .smorest import (  # noqa
    Api,
    Blueprint,
    Schema,
    AutoSchema,
    SQLCursorPage,
    SortField,
)
from .integrity_error import catch_integrity_error  # noqa
